﻿using System;
using System.Collections.Generic;
using System.Linq;
using Maze.Input.Events;
using Maze.Input.Wrapper;
using Microsoft.Xna.Framework.Input;
using FRBInputManager = FlatRedBall.Input.InputManager;
using Mouse = FlatRedBall.Input.Mouse;

namespace Maze.Input
{
    public class InputManager
    {
        private static InputManager _instance;

        public event EventHandler DetectedInputChanged;

        private readonly InputDetectedEvent _inputDetectedEvent;
        private readonly WrapperFactory _wrapperFactory;
        private Input _detectedInput;

        protected InputManager()
        {
            _inputDetectedEvent = new InputDetectedEvent();
            _wrapperFactory = new WrapperFactory();

            DetectedInput = Input.None;

            SubscribeToEvents();
        }

        public static InputManager GetInstance()
        {
            return _instance ?? (_instance = new InputManager());
        }

        public Mouse Mouse
        {
            get
            {
                return FRBInputManager.Mouse;
            }
        }

        public Input DetectedInput
        {
            get
            {
                return _detectedInput;
            }
            private set
            {
                var oldInput = _detectedInput;

                _detectedInput = value;

                if (oldInput != _detectedInput)
                {
                    GetInputWrappers(value);
                }
            }
        }

        public ReadOnlyWrapperList Wrappers
        {
            get;
            private set;
        }

        public bool HasPollableWrapper
        {
            get
            {
                return Wrappers.PollableWrappers.Count > 0;
            }
        }

        public bool HasUnpollableWrapper
        {
            get
            {
                return Wrappers.UnpollableWrappers.Count > 0;
            }
        }

        public void Destroy()
        {
            UnsubscribeFromEvents();

            foreach (var wrapper in Wrappers)
            {
                wrapper.Destroy();
            }

            InputDetecter.GetInstance().Destroy();

            _instance = null;
        }

        public void DetectInput()
        {
            DetectedInput = InputDetecter.GetInstance().DetectInputs();
            OnDetectedInputChanged();
        }

        public bool KeyDown(Keys key)
        {
            return Wrappers.Where(wrapper => wrapper.IsActive).Any(wrapper => wrapper.KeyDown(key));
        }

        public bool KeyPushed(Keys key)
        {
            return Wrappers.Where(wrapper => wrapper.IsActive).Any(wrapper => wrapper.KeyPushed(key));
        }

        public List<string> GetErrors()
        {
            var errorList = new List<string>();
            foreach (var errors in Wrappers.Select(wrapper => wrapper.Errors).Where(errors => errors.Count > 0))
            {
                errorList.AddRange(errors);
            }
            return errorList;
        }

        public TType GetInputWrapper<TType>() where TType : BaseInputWrapper
        {
            return (TType) Wrappers.GetWrapper(typeof(TType));
        }

        private void GetInputWrappers(Input inputs)
        {
            Wrappers = _wrapperFactory.GetWrappers(inputs);
        }

        private void OnInputDetected(object eventArgs)
        {
            DetectInput();
        }

        private void OnDetectedInputChanged()
        {
            var handler = DetectedInputChanged;
            if (handler != null) handler(null, EventArgs.Empty);
        }

        private void SubscribeToEvents()
        {
            _inputDetectedEvent.Subscribe(OnInputDetected);
        }

        private void UnsubscribeFromEvents()
        {
            _inputDetectedEvent.Unsubscribe(OnInputDetected);
        }
    }
}
