﻿using System;
using System.Collections.Generic;
using System.Linq;
using FlatRedBall;
using Kinect.Toolbox;
using Maze.Input.Kinect.Events;
using Microsoft.Research.Kinect.Nui;
using Microsoft.Xna.Framework.Graphics;
using Vector2 = Microsoft.Xna.Framework.Vector2;
using Vector3 = Microsoft.Xna.Framework.Vector3;

namespace Maze.Input.Kinect.Models
{
    public class Kinect
    {
        private const int MaxTilt = 27;
        private const int MinTilt = -MaxTilt;
        private const float HandInnerBound = .2f;
        private const float HandOuterBound = .3f;
        private const float MinHandBodyDistance = .5f;
        private const float Threshold = 20;
        private const string VideoNotWorkingErrorText = "Kinect error: Not streaming video";
        private const string SkeletonNotWorkingErrorText = "Kinect error: Not detecting skeletons";
        private const string NoSkeletonFoundErrorText = "No player detected";
        private const string HandsTooCloseErrorText = "Your hands are too close to eachother";
        private const string HandsTooFarAwayErrorText = "Your hands are too far away from eachother";
        private const string HandTooCloseToBodyErrorText = "Your hands are too close to your body";

        public delegate void ActionDetectedEventHandler(Kinect kinect, Actions action);

        public delegate void KinectActivityEventHandler(Kinect kinect);

        public event ActionDetectedEventHandler ActionDetected;
        public event KinectActivityEventHandler Activity;

        public enum Actions
        {
            SwipeLeft,
            SwipeRight,
            LeftHandOverHead,
            RightHandOverHead,
            LeftHello,
            RightHello
        }

        private readonly SkeletonProcessedEvent _skeletonProcessedEvent;
        private readonly VideoProcessedEvent _videoProcessedEvent;
        private readonly SwipeGestureDetector _gestureDetector;
        private readonly AlgorithmicPostureDetector _postureDetector;
        private bool _tiltInitialized;
        private Vector3 _lastLeftHandReading;
        private Vector3 _lastRightHandReading;

        public Kinect(Runtime kinectRuntime)
        {
            Runtime = kinectRuntime;
            _skeletonProcessedEvent = new SkeletonProcessedEvent();
            _videoProcessedEvent = new VideoProcessedEvent();
            _gestureDetector = new SwipeGestureDetector();
            _postureDetector = new AlgorithmicPostureDetector();

            Errors = new List<string>();
            GameplayAllowed = false;

            AddError(VideoNotWorkingErrorText);
            AddError(SkeletonNotWorkingErrorText);

            SubscribeToEvents();
            InitializeKinect();
        }

        public Runtime Runtime
        {
            get;
            private set;
        }

        public float Angle
        {
            get;
            set;
        }

        public int CameraTilt
        {
            get
            {
                try
                {
                    return Runtime.NuiCamera.ElevationAngle;
                }
                catch (Exception)
                {
                    return 0;
                }
            }
            set
            {
                if (value > MaxTilt || value < MinTilt)
                {
                    value = value > MaxTilt ? MaxTilt : MinTilt;
                }

                try
                {
                    Runtime.NuiCamera.ElevationAngle = value;
                    var cameraTiltAdjustedEvent = new CameraTiltAdjustedEvent();
                    cameraTiltAdjustedEvent.Publish(new CameraTiltAdjustedEventArgs { Tilt = value });
                }
                catch (Exception)
                {
                    Console.WriteLine("Catched an exception while adjusting the Camera Tilt");
                }
            }
        }

        public Texture2D VideoFrame
        {
            get;
            private set;
        }

        public List<string> Errors
        {
            get;
            private set;
        }

        public bool GameplayAllowed
        {
            get;
            private set;
        }

        public void Destroy()
        {
            UnsubscribeFromEvents();
            Runtime.Uninitialize();
        }

        private void InitializeKinect()
        {
            Runtime.Initialize(RuntimeOptions.UseColor | RuntimeOptions.UseSkeletalTracking | RuntimeOptions.UseDepthAndPlayerIndex);

            Runtime.SkeletonEngine.TransformSmooth = true;
//            var parameters = new TransformSmoothParameters
//                                 {
//                                     Smoothing = 0.7f,
//                                     Correction = 0.3f,
//                                     Prediction = 0.4f,
//                                     JitterRadius = 1.0f,
//                                     MaxDeviationRadius = 0.5f
//                                 };

            var parameters = new TransformSmoothParameters
            {
                Smoothing = 0.75f,
                Correction = 0.2f,
                Prediction = 0.2f,
                JitterRadius = .05f,
                MaxDeviationRadius = 0.05f
            };

//            var parameters = new TransformSmoothParameters
//            {
//                Smoothing = 1.0f,
//                Correction = 0.1f,
//                Prediction = 0.1f,
//                JitterRadius = 0.05f,
//                MaxDeviationRadius = 0.05f
//            };
            Runtime.SkeletonEngine.SmoothParameters = parameters;

            Runtime.VideoStream.Open(ImageStreamType.Video, 2, ImageResolution.Resolution640x480, ImageType.Color);
        }

        private void InitializeTilt()
        {
            if (_tiltInitialized) return;
            CameraTilt = CameraTilt;
            _tiltInitialized = true;
        }

        private void SubscribeToEvents()
        {
            Runtime.SkeletonFrameReady += SkeletonFrameReady;
            Runtime.VideoFrameReady += VideoFrameReady;

            _gestureDetector.OnGestureDetected += OnGestureDetected;
            _postureDetector.PostureDetected += OnPostureDetected;

            var cameraTiltEvent = new AdjustCameraTiltEvent();
            cameraTiltEvent.Subscribe(OnAdjustCameraTilt);
        }

        private void UnsubscribeFromEvents()
        {
            Runtime.SkeletonFrameReady -= SkeletonFrameReady;
            Runtime.VideoFrameReady -= VideoFrameReady;

            _gestureDetector.OnGestureDetected -= OnGestureDetected;
            _postureDetector.PostureDetected -= OnPostureDetected;

            var cameraTiltEvent = new AdjustCameraTiltEvent();
            cameraTiltEvent.Unsubscribe(OnAdjustCameraTilt);
        }

        private void VideoFrameReady(object sender, ImageFrameReadyEventArgs e)
        {
            RemoveError(VideoNotWorkingErrorText);
            InitializeTilt();

            var imageData = e.ImageFrame.Image;

            if (VideoFrame == null)
            {
                VideoFrame = new Texture2D(FlatRedBallServices.GraphicsDevice, imageData.Width, imageData.Height, false, SurfaceFormat.Color);
            }

            var imageBits = imageData.Bits;

            // Flip the R with the B
            for (int b = 0; b < imageBits.Length; b += 4)
            {
                byte tmp = imageBits[b];
                imageBits[b] = imageBits[b + 2];
                imageBits[b + 2] = tmp;
            }
            VideoFrame.SetData(imageBits);

            _videoProcessedEvent.Publish(new VideoProcessedEventArgs { Frame = VideoFrame });
        }

        private void SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            RemoveError(SkeletonNotWorkingErrorText);
            InitializeTilt();

            if (e.SkeletonFrame == null)
                return;

            var skeletons = e.SkeletonFrame.Skeletons.Where(s => s.TrackingState == SkeletonTrackingState.Tracked);

            if (skeletons.Count() == 0)
            {
                GameplayAllowed = false;
                AddError(NoSkeletonFoundErrorText);

                _skeletonProcessedEvent.Publish(new SkeletonProcessedEventArgs { Skeleton = null, Kinect = this });
                return;
            }
            GameplayAllowed = true;
            RemoveError(NoSkeletonFoundErrorText);

            foreach (var skeleton in skeletons)
            {
                ProcessSkeleton(skeleton);
                _skeletonProcessedEvent.Publish(new SkeletonProcessedEventArgs { Skeleton = skeleton, Kinect = this });
            }
        }

        private void ProcessSkeleton(SkeletonData skeleton)
        {
            var leftHand = skeleton.Joints[JointID.HandLeft];
            var rightHand = skeleton.Joints[JointID.HandRight];

            _gestureDetector.Add(rightHand.Position, Runtime.SkeletonEngine);

            if (leftHand.TrackingState != JointTrackingState.Tracked || rightHand.TrackingState != JointTrackingState.Tracked)
            {
                return;
            }

            if (DoesExceedThreshold(leftHand, rightHand))
            {
                return;
            }

            CalculateAngle(leftHand, rightHand);
            CalculateDistanceBetweenHands(leftHand, rightHand);
            HandleHandsToBodyDistance(leftHand, rightHand, skeleton);

            OnActivity();

            _postureDetector.TrackPostures(skeleton);
        }

        private bool DoesExceedThreshold(Joint leftHand, Joint rightHand)
        {
            var leftHandPosition = ToVector(leftHand.Position);
            var rightHandPosition = ToVector(rightHand.Position);

            if (_lastLeftHandReading.Y == 0)
            {
                _lastLeftHandReading = leftHandPosition;
            }
            if (_lastRightHandReading.Y == 0)
            {
                _lastRightHandReading = rightHandPosition;
            }

            var exceedsThreshold = ExceedsThreshold(leftHand, _lastLeftHandReading) ||
                                   ExceedsThreshold(rightHand, _lastRightHandReading);
            
            _lastLeftHandReading = leftHandPosition;
            _lastRightHandReading = rightHandPosition;
            
            return exceedsThreshold;
        }

        private Vector3 ToVector(Vector vector)
        {
            return new Vector3(vector.X, vector.Y, vector.Z);
        }

        private bool ExceedsThreshold(Joint hand, Vector3 lastReading)
        {
            return Math.Abs(lastReading.X - hand.Position.X) > Threshold && Math.Abs(lastReading.Y - hand.Position.Y) > Threshold;
        }

        private void CalculateAngle(Joint leftHand, Joint rightHand)
        {
            var yDifference = rightHand.Position.Y - leftHand.Position.Y;
            var xDifference = rightHand.Position.X - leftHand.Position.X;

            var angle = Math.Atan2(yDifference, xDifference) * (180 / Math.PI);
            Angle = (float) angle * -1;
        }

        private void CalculateDistanceBetweenHands(Joint leftHand, Joint rightHand)
        {
            var centerPoint = CalculateCenter(leftHand, rightHand);

            var isLeftHandInsideInnerBound = IsInsideCirle(leftHand, centerPoint, HandInnerBound);
            var isRightHandInsideInnerBound = IsInsideCirle(rightHand, centerPoint, HandInnerBound);
            if (isLeftHandInsideInnerBound && isRightHandInsideInnerBound)
            {
                // Hands are too close
                GameplayAllowed = false;
                AddError(HandsTooCloseErrorText);
                return;
            }
            RemoveError(HandsTooCloseErrorText);

            var isLeftHandInsideOuterBound = IsInsideCirle(leftHand, centerPoint, HandOuterBound);
            var isRightHandInsideOuterBound = IsInsideCirle(rightHand, centerPoint, HandOuterBound);
            if (!isLeftHandInsideOuterBound && !isRightHandInsideOuterBound)
            {
                // Hands are too far away
                GameplayAllowed = false;
                AddError(HandsTooFarAwayErrorText);
                return;
            }
            RemoveError(HandsTooFarAwayErrorText);
        }

        private void HandleHandsToBodyDistance(Joint leftHand, Joint rightHand, SkeletonData skeleton)
        {
            var shoulderCenter = skeleton.Joints[JointID.ShoulderCenter];
            var leftShoulder = skeleton.Joints[JointID.ShoulderLeft];
            var leftElbow = skeleton.Joints[JointID.ElbowLeft];

            var armLength = MeasureOverArmLength(leftShoulder, leftElbow) + MeasureUnderArmLength(leftHand, leftElbow);
            var minimumDistance = MinHandBodyDistance * armLength;

            var leftHandDistanceAcceptable = (shoulderCenter.Position.Z - leftHand.Position.Z) >= minimumDistance;
            if (!leftHandDistanceAcceptable)
            {
                GameplayAllowed = false;
                AddError(HandTooCloseToBodyErrorText);
                return;
            }

            var rightHandDistanceAcceptable = (shoulderCenter.Position.Z - rightHand.Position.Z) >= minimumDistance;
            if (!rightHandDistanceAcceptable)
            {
                GameplayAllowed = false;
                AddError(HandTooCloseToBodyErrorText);
                return;
            }

            RemoveError(HandTooCloseToBodyErrorText);
        }

        private double MeasureOverArmLength(Joint leftShoulder, Joint leftElbow)
        {
            return CalculateBodyPartLength(leftShoulder, leftElbow);
        }

        private double MeasureUnderArmLength(Joint leftHand, Joint leftElbow)
        {
            return CalculateBodyPartLength(leftHand, leftElbow);
        }

        private double CalculateBodyPartLength(Joint first, Joint second)
        {
            var vector = new Vector3(
                Math.Max(first.Position.X, second.Position.X) - Math.Min(first.Position.X, second.Position.X),
                Math.Max(first.Position.Y, second.Position.Y) - Math.Min(first.Position.Y, second.Position.Y),
                Math.Max(first.Position.Z, second.Position.Z) - Math.Min(first.Position.Z, second.Position.Z)
                );

            return Math.Sqrt(vector.X * vector.X + vector.Y * vector.Y + vector.Z * vector.Z);
        }

        private void AddError(string errorText)
        {
            if (Errors.Contains(errorText)) return;
            Errors.Add(errorText);
        }

        private void RemoveError(string errorText)
        {
            if (!Errors.Contains(errorText)) return;
            Errors.Remove(errorText);
        }

        private void OnAdjustCameraTilt(AdjustCameraTiltEventArgs args)
        {
            if (args.Absolute)
            {
                CameraTilt = args.Tilt;
                return;
            }
            CameraTilt += args.Tilt;
        }

        private void OnActivity()
        {
            var handler = Activity;
            if (handler != null) handler(this);
        }

        private void OnGestureDetected(string gestureString)
        {
            HandleUserAction(gestureString);
        }

        private void OnPostureDetected(string postureString)
        {
            HandleUserAction(postureString);
        }

        private void HandleUserAction(string action)
        {
            Actions gesture;
            if (action.Equals("SwipeToLeft"))
            {
                gesture = Actions.SwipeLeft;
            }
            else if (action.Equals("SwipeToRight"))
            {
                gesture = Actions.SwipeRight;
            }
            else if (action.Equals("LeftHandOverHead"))
            {
                gesture = Actions.LeftHandOverHead;
            }
            else if (action.Equals("RightHandOverHead"))
            {
                gesture = Actions.RightHandOverHead;
            }
            else if (action.Equals("LeftHello"))
            {
                gesture = Actions.LeftHello;
            }
            else if (action.Equals("RightHello"))
            {
                gesture = Actions.RightHello;
            }
            else
            {
                return;
            }

            var handler = ActionDetected;
            if (handler != null) handler(this, gesture);
        }

        private static Vector2 CalculateCenter(Joint leftHand, Joint rightHand)
        {
            float xCenter;
            float yCenter;

            if (leftHand.Position.X > rightHand.Position.X)
            {
                xCenter = leftHand.Position.X - (leftHand.Position.X - rightHand.Position.X) / 2;
            }
            else
            {
                xCenter = rightHand.Position.X - (rightHand.Position.X - leftHand.Position.X) / 2;
            }

            if (leftHand.Position.Y > rightHand.Position.Y)
            {
                yCenter = leftHand.Position.Y - (leftHand.Position.Y - rightHand.Position.Y) / 2;
            }
            else
            {
                yCenter = rightHand.Position.Y - (rightHand.Position.Y - leftHand.Position.Y) / 2;
            }
            return new Vector2(xCenter, yCenter);
        }

        private static bool IsInsideCirle(Joint joint, Vector2 centerPoint, float radius)
        {
            var distance = Math.Pow(centerPoint.X - joint.Position.X, 2) + Math.Pow(centerPoint.Y - joint.Position.Y, 2);
            return distance < Math.Pow(radius, 2);
        }
    }
}
