﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.Xna.Framework;
using PSMoveSharp;

namespace Maze.Input.Move.Models
{
    public class Move
    {
        public enum Buttons
        {
            Square,
            Cross,
            Circle,
            Triangle,
            Move,
            Start,
            Select
        }

        public delegate void MoveActivityEventHandler(Move move);
        public event MoveActivityEventHandler Activity;

        private readonly MoveClient _moveClient;
        private readonly Dictionary<Buttons, bool> _pushedButtons;
        private readonly Dictionary<Buttons, DateTime> _quarantine;
        private readonly TimeSpan _quarantineAmount;

        public Move(MoveClient.MoveID moveID)
        {
            MoveID = moveID;

            _moveClient = MoveClient.GetInstance();
            _pushedButtons = new Dictionary<Buttons, bool>();
            _quarantine = new Dictionary<Buttons, DateTime>();
            _quarantineAmount = new TimeSpan(0, 0, 0, 0, 200);

            InitializePushedButtons();
            SubscribeToEvents();
        }

        public MoveClient.MoveID MoveID
        {
            get;
            private set;
        }

        public bool Calibrated
        {
            get;
            private set;
        }

        public Vector3 Position
        {
            get;
            private set;
        }

        public Vector3 Orientation
        {
            get;
            private set;
        }

        public Vector3 HandlePosition
        {
            get;
            private set;
        }

        public void Reset()
        {
            Calibrated = false;
            _moveClient.ResetController(TranslateToGemID(MoveID));
        }

        public void Calibrate()
        {
            if (Calibrated) return;
            var calibrateTask = new Task(PerformCalibration);
            calibrateTask.Start();
        }

        public bool ButtonPushed(Buttons button)
        {
            var isButtonPushed = _pushedButtons[button];
            if (isButtonPushed)
            {
                var isQuarantined = _quarantine.ContainsKey(button);
                if (isQuarantined)
                {
                    var shouldRemoveFromQuarantine = DateTime.Now - _quarantine[button] >= _quarantineAmount;
                    if (shouldRemoveFromQuarantine)
                    {
                        _quarantine.Remove(button);
                        isQuarantined = false;
                    }
                }

                if (!isQuarantined)
                {
                    _quarantine.Add(button, DateTime.Now);
                    return true;
                }
            }
            return false;
        }

        public bool ButtonDown(Buttons button)
        {
            return _pushedButtons[button];
        }

        public void Destroy()
        {
            UnsubscribeFromEvents();
        }

        private void PerformCalibration()
        {
            _moveClient.CalibrateController(TranslateToGemID(MoveID));
            Thread.Sleep(1000);
            _moveClient.TrackHues();
        }

        private int TranslateToGemID(MoveClient.MoveID moveID)
        {
            return (int) moveID >> 1;
        }

        private void OnMoveStateUpdated(PSMoveSharpState state)
        {
            var gemId = TranslateToGemID(MoveID);

            var moveStatus = state.gemStatus[gemId];
            var moveState = state.gemStates[gemId];
            
            SetPushedButtons(moveState.pad);
            SetCalibrated(moveStatus);

            if (!Calibrated)
            {
                return;
            }

            Position = CalculatePosition(moveState);
            Orientation = ConvertToOrientation(moveState.quat);
            HandlePosition = new Vector3(moveState.handle_pos.x, moveState.handle_pos.y, moveState.handle_pos.z);

            OnActivity();
        }

        private void SetPushedButtons(PSMoveSharpPadData pad)
        {
            var buttons = pad.digitalbuttons;

            _pushedButtons[Buttons.Circle] = (buttons & PSMoveSharpConstants.ctrlCircle) != 0;
            _pushedButtons[Buttons.Cross] = (buttons & PSMoveSharpConstants.ctrlCross) != 0;
            _pushedButtons[Buttons.Move] = (buttons & PSMoveSharpConstants.ctrlTick) != 0;
            _pushedButtons[Buttons.Select] = (buttons & PSMoveSharpConstants.ctrlSelect) != 0;
            _pushedButtons[Buttons.Square] = (buttons & PSMoveSharpConstants.ctrlSquare) != 0;
            _pushedButtons[Buttons.Start] = (buttons & PSMoveSharpConstants.ctrlStart) != 0;
            _pushedButtons[Buttons.Triangle] = (buttons & PSMoveSharpConstants.ctrlTriangle) != 0;
        }

        private void SetCalibrated(PSMoveSharpStatus moveStatus)
        {
            Calibrated = !(moveStatus.connected == 0 && moveStatus.status_code == 2);
        }

        private Vector3 CalculatePosition(PSMoveSharpGemState state)
        {
            return new Vector3(state.pos.x, state.pos.y, state.pos.z);
        }

        private Vector3 ConvertToOrientation(Float4 quat)
        {
            Vector3 orientation;
            orientation.X = Convert.ToInt32((180.0 / Math.PI) * QuaternionToEuler(quat).X);
            orientation.Y = Convert.ToInt32((180.0 / Math.PI) * QuaternionToEuler(quat).Y);
            orientation.Z = Convert.ToInt32((180.0 / Math.PI) * QuaternionToEuler(quat).Z);
            return orientation;
        }

        public Vector3 QuaternionToEuler(Float4 quat)
        {
            Vector3 euler;

            euler.Y = Convert.ToSingle(Math.Asin(2.0 * ((quat.x * quat.z) - (quat.w * quat.y))));

            if (euler.Y == 90.0)
            {
                euler.X = Convert.ToSingle(2.0 * Math.Atan2(quat.x, quat.w));
                euler.Z = 0;
            }
            else if (euler.Y == -90.0)
            {
                euler.X = Convert.ToSingle(-2.0 * Math.Atan2(quat.x, quat.w));
                euler.Z = 0;
            }
            else
            {
                euler.X = Convert.ToSingle(Math.Atan2(2.0 * ((quat.x * quat.y) + (quat.z * quat.w)), 1.0 - (2.0 * ((quat.y * quat.y) + (quat.z * quat.z)))));
                euler.Z = Convert.ToSingle(Math.Atan2(2.0 * ((quat.x * quat.w) + (quat.y * quat.z)), 1.0 - (2.0 * ((quat.z * quat.z) + (quat.w * quat.w)))));
            }
            return euler;
        }

        private void OnActivity()
        {
            var handler = Activity;
            if (handler != null) handler(this);
        }

        private void SubscribeToEvents()
        {
            _moveClient.MoveStateUpdated += OnMoveStateUpdated;
        }

        private void UnsubscribeFromEvents()
        {
            _moveClient.MoveStateUpdated -= OnMoveStateUpdated;
        }

        private void InitializePushedButtons()
        {
            foreach (Buttons button in Enum.GetValues(typeof(Buttons)))
            {
                _pushedButtons.Add(button, false);
            }
        }
    }
}
