﻿using System;
using System.Net;
using PSMoveSharp;

namespace Maze.Input.Move
{
    public class MoveClient
    {
        private static MoveClient _instance;

        [Flags]
        public enum MoveID
        {
            None = 0,
            First = 1,
            Second = 1 << 1,
            Third = 1 << 2,
            Fourth = 1 << 3
        }

        public enum ConnectionStatus
        {
            Connected,
            Disconnected
        }

        public delegate void MoveStateUpdatedEventHandler(PSMoveSharpState state);
        public delegate void CameraFrameStateUpdatedEventHandler(PSMoveSharpCameraFrameState state);
        public delegate void ConnectedMovesUpdatedEventHandler(MoveID moveID, ConnectionStatus connectionStatus);

        public event MoveStateUpdatedEventHandler MoveStateUpdated;
        public event CameraFrameStateUpdatedEventHandler CameraFrameStateUpdated;

        public event ConnectedMovesUpdatedEventHandler ConnectedMovesUpdated;

        private readonly MoveClientDelegate _client;

        private int _delay;

        protected MoveClient()
        {
            _client = new MoveClientDelegate() { StateUpdated = StateUpdated, CameraStateUpdated = CameraStateUpdated };
            _delay = 16;
        }

        public static MoveClient GetInstance()
        {
            return _instance ?? (_instance = new MoveClient());
        }

        public bool IsConnected
        {
            get;
            private set;
        }

        public bool IsPaused
        {
            get;
            private set;
        }

        public MoveID ConnectedMoves
        {
            get;
            private set;
        }

        public int Delay
        {
            get
            {
                return _delay;
            }
            set
            {
                _delay = ChangeDelay(value);
            }
        }

        public void Connect()
        {
            if (!Globals.PSMove) return;
            Connect(Globals.PSMoveServerAddr, Globals.PSMoveServerPort);
        }

        public void Connect(String serverAddress, int serverPort)
        {
            if (IsConnected) return;

            var server = Dns.GetHostAddresses(serverAddress)[0].ToString();
            var timeout = TimeSpan.FromMilliseconds(Globals.PSMoveConnectTimeoutMs);

            try
            {
                _client.ConnectWithTimeout(server, serverPort, timeout);
                _client.StartThread();

                IsConnected = true;
            }
            catch(TimeoutException)
            {
                return;
            }
        }

        public void Disconnect()
        {
            if (!IsConnected) return;

            try
            {
                _client.StopThread();
                _client.Close();

                IsConnected = false;
            }
            catch
            {
                return;
            }
        }

        public void ResetController(int gemId)
        {
            _client.SendRequestPacket(PSMoveClient.ClientRequest.PSMoveClientRequestControllerReset, Convert.ToUInt32(gemId));
        }

        public void CalibrateController(int gemId)
        {
            _client.CalibrateController(gemId);
        }

        public void TrackHues()
        {
            _client.TrackAllHues();
        }

        public void PauseTracking()
        {
            _client.Pause();
            IsPaused = true;
        }

        public void ResumeTracking()
        {
            _client.Resume();
            IsPaused = false;
        }

        public void Destroy()
        {
            if (IsConnected)
            {
                Disconnect();
            }
            _instance = null;
        }

        private int ChangeDelay(int msDelay)
        {
            var delay = Convert.ToUInt32(msDelay);
            delay = Math.Max(delay, 1);
            delay = Math.Min(delay, 255);
            _client.DelayChange(delay);
            
            return Convert.ToInt32(delay);
        }

        private void StateUpdated(PSMoveSharpState state)
        {
            UpdateConnectedMoves(state);
            OnMoveStateUpdated(state);
        }

        private void CameraStateUpdated(PSMoveSharpCameraFrameState state)
        {
            OnCameraStateUpdated(state);
        }

        private void UpdateConnectedMoves(PSMoveSharpState state)
        {
            for (int index = 0; index < state.gemStatus.Length; index++)
            {
                var moveId = (MoveID) (1 << index);
                var gem = state.gemStatus[index];

                var isConnected = gem.connected == 1;
                var isReset = gem.status_code == 2;
                var isCalibrated = gem.status_code == 3;
                var isCalibrating = gem.status_code == 4;

                var shouldAdd = isConnected || isReset || isCalibrated || isCalibrating;
                var shouldRemove = !isConnected && gem.status_code == 1;

                if (shouldAdd && !ConnectedMoves.HasFlag(moveId))
                {
                    AddMove(moveId);
                }
                else if (shouldRemove && ConnectedMoves.HasFlag(moveId))
                {
                    RemoveMove(moveId);
                }
            }
        }

        private void AddMove(MoveID moveId)
        {
            ConnectedMoves = ConnectedMoves | moveId;
            OnConnectedMovesUpdated(moveId, ConnectionStatus.Connected);
        }

        private void RemoveMove(MoveID moveId)
        {
            ConnectedMoves = ConnectedMoves ^ moveId;
            OnConnectedMovesUpdated(moveId, ConnectionStatus.Disconnected);
        }

        private void OnConnectedMovesUpdated(MoveID moveId, ConnectionStatus status)
        {
            var handler = ConnectedMovesUpdated;
            if (handler != null) handler(moveId, status);
        }

        private void OnMoveStateUpdated(PSMoveSharpState state)
        {
            var handler = MoveStateUpdated;
            if (handler != null) handler(state);
        }

        public void OnCameraStateUpdated(PSMoveSharpCameraFrameState state)
        {
            var handler = CameraFrameStateUpdated;
            if (handler != null) handler(state);
        }
    }
}
