﻿using System;
using System.Drawing;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using PSMoveSharp;

namespace Maze.Input.Move
{
    class MoveClientDelegate : MoveClientExtension
    {
        public delegate void StateUpdatedDelegate(PSMoveSharpState state);
        public delegate void CameraStateUpdatedDelegate(PSMoveSharpCameraFrameState state);

        public StateUpdatedDelegate StateUpdated;
        public CameraStateUpdatedDelegate CameraStateUpdated;

        private readonly PSMoveSharpState _latestState;
        private readonly PSMoveSharpCameraFrameState _latestCameraFrameState;
        private readonly Thread _readerThread;
        private bool _readerThreadShuttingDown;
        private bool _reading;

        public MoveClientDelegate()
        {
            _readerThread = new Thread(DoReceive);
            _readerThreadShuttingDown = false;

            _latestState = new PSMoveSharpState();
            _latestCameraFrameState = new PSMoveSharpCameraFrameState();

            _reading = false;
        }

        /* Destructor */
        ~MoveClientDelegate()
        {
            ShutdownThread();
        }

        public void StartThread()
        {
            _readerThread.Start();
        }

        public void StopThread()
        {
            ShutdownThread();
        }

        private void ShutdownThread()
        {
            _readerThreadShuttingDown = true;

            try
            {
                _readerThread.Join();
            }
            catch
            {
                return;
            }
        }

        private void DoReceive()
        {
            var remoteEndPoint = new IPEndPoint(IPAddress.Any, 0);
            var udpState = new UdpState();

            while (!_readerThreadShuttingDown)
            {
                udpState.IpEndPoint = remoteEndPoint;
                udpState.UdpClient = _udpClient;

                _reading = true;
                try
                {
                    if (_udpClient != null)
                    {
                        _udpClient.BeginReceive(OnReceiveData, udpState);
                    }
                }
                catch
                {
                    Console.WriteLine("Exception from UDP-client beginReceive");
                }

                while (_reading)
                {
                    if (_readerThreadShuttingDown)
                    {
                        break;
                    }
                    Thread.Sleep(0);
                }
            }
        }

        protected void OnReceiveData(IAsyncResult asyncResult)
        {
            var udpState = (UdpState) asyncResult.AsyncState;
            var udpClient = udpState.UdpClient;
            var ipEndPoint = udpState.IpEndPoint;

            Byte[] buffer;
            try
            {
                buffer = udpClient.EndReceive(asyncResult, ref ipEndPoint);
            }
            catch (ObjectDisposedException)
            {
                _reading = false;
                return;
            }
            catch (SocketException)
            {
                _reading = false;
                return;
            }

            var packetLength = NetworkReaderHelper.ReadInt32(ref buffer, 12);
            var magic = NetworkReaderHelper.ReadInt32(ref buffer, 0);
            var code = NetworkReaderHelper.ReadInt32(ref buffer, 8);
            var packetIndex = NetworkReaderHelper.ReadUint32(ref buffer, 16);

            switch (code)
            {
                case 1:
                    ProcessStandardStatePacket(buffer);
                    break;
//                case 2:
//                    ProcessCameraFrameSlicePacket(buffer, packetLength, packetIndex);
//                    break;
//                case 3:
//                    ProcessCameraFrameStatePacket(buffer, packetIndex);
//                    break;
            }
            _reading = false;
        }

        private void ProcessStandardStatePacket(byte[] buffer)
        {
            _latestState.packet_index = NetworkReaderHelper.ReadUint32(ref buffer, 16);
            _latestState.serverConfig.FillFromNetworkBuffer(ref buffer);

            for (int i = 0; i < PSMoveSharpState.PSMoveSharpNumMoveControllers; i++)
            {
                _latestState.gemStatus[i].FillFromNetworkBuffer(ref buffer, i);
                _latestState.gemStates[i].FillFromNetworkBuffer(ref buffer, i);
                _latestState.imageStates[i].FillFromNetworkBuffer(ref buffer, i);
                _latestState.pointerStates[i].FillFromNetworkBuffer(ref buffer, i);
                _latestState.sphereStates[i].FillFromNetworkBuffer(ref buffer, i);
                _latestState.positionPointerStates[i].FillFromNetworkBuffer(ref buffer, i);
            }
            
            _latestState.navInfo.FillFromNetworkBuffer(ref buffer);
            
            for (int i = 0; i < PSMoveSharpNavInfo.PSMoveSharpNumNavControllers; i++)
            {
                _latestState.padData[i].FillFromNetworkBuffer(ref buffer, i);
            }

            if (StateUpdated != null)
            {
                StateUpdated(_latestState);
            }
        }

        private void ProcessCameraFrameSlicePacket(byte[] buffer, int packetLength, uint packetIndex)
        {
            var jpegData = new byte[packetLength - 3];
            int sliceNumber = NetworkReaderHelper.ReadByte(ref buffer, 20);
            int numberOfSlices = NetworkReaderHelper.ReadByte(ref buffer, 21);
            int format = NetworkReaderHelper.ReadByte(ref buffer, 22);
            
            var rowHeight = 480 / numberOfSlices;
            var rowStart = rowHeight * sliceNumber;

            Array.Copy(buffer, 23, jpegData, 0, packetLength - 3);

//            var jpegStream = new MemoryStream(jpegData);
//            Image slice;
//            try
//            {
//                slice = Image.FromStream(jpegStream);
//            }
//            catch
//            {
//                return;
//            }
//            _latestCameraFrameState.ProcessPacket((int) packetIndex, slice, rowStart, rowHeight);

            using (var jpegStream = new MemoryStream(jpegData))
            {
                Image slice;
                try
                {
                    slice = Image.FromStream(jpegStream);
                }
                catch
                {
                    return;
                }
                _latestCameraFrameState.ProcessPacket((int)packetIndex, slice, rowStart, rowHeight);
            }

            if (CameraStateUpdated != null)
            {
                CameraStateUpdated(_latestCameraFrameState);
            }
        }

        private void ProcessCameraFrameStatePacket(byte[] buffer, uint packetIndex)
        {
            var cameraState = new PSMoveSharpState();

            cameraState.packet_index = NetworkReaderHelper.ReadUint32(ref buffer, 16);
            cameraState.serverConfig.FillFromNetworkBuffer(ref buffer);

            for (int i = 0; i < PSMoveSharpState.PSMoveSharpNumMoveControllers; i++)
            {
                cameraState.gemStatus[i].FillFromNetworkBuffer(ref buffer, i);
                cameraState.gemStates[i].FillFromNetworkBuffer(ref buffer, i);
                cameraState.imageStates[i].FillFromNetworkBuffer(ref buffer, i);
                cameraState.pointerStates[i].FillFromNetworkBuffer(ref buffer, i);
                cameraState.sphereStates[i].FillFromNetworkBuffer(ref buffer, i);
                cameraState.positionPointerStates[i].FillFromNetworkBuffer(ref buffer, i);
            }

            cameraState.navInfo.FillFromNetworkBuffer(ref buffer);

            for (int i = 0; i < PSMoveSharpNavInfo.PSMoveSharpNumNavControllers; i++)
            {
                cameraState.padData[i].FillFromNetworkBuffer(ref buffer, i);
            }

            _latestCameraFrameState.ProcessStatePacket((int) packetIndex, cameraState);
        }

        public void SetNumImageSlices(int slices)
        {
            _latestCameraFrameState.camera_frame_state_collector.SetNumSlices(slices);
        }

        private sealed class UdpState
        {
            public IPEndPoint IpEndPoint
            {
                get;
                set;
            }

            public UdpClient UdpClient
            {
                get;
                set;
            }
        }
    }
}
