﻿using System;
using System.Linq;
using Maze.Input.Events;
using Maze.Input.Move.Events;
using Maze.Input.Move.Models;

namespace Maze.Input.Move
{
    class MoveDetecter
    {
        private static MoveDetecter _instance;
        
        private readonly MoveClient _moveClient;
        private readonly InputDetectedEvent _inputDetectedEvent;

        protected MoveDetecter()
        {
            _moveClient = MoveClient.GetInstance();
            _inputDetectedEvent = new InputDetectedEvent();

            Moves = new MoveCollection();

            SubscribeToEvents();
            DetectMoves();
        }

        public static MoveDetecter GetInstance()
        {
            return _instance ?? (_instance = new MoveDetecter());
        }

        public bool Connected
        {
            get
            {
                return _moveClient.IsConnected;
            }
        }

        public MoveCollection Moves
        {
            get;
            private set;
        }

        public int MoveCount
        {
            get
            {
                return Moves.Count;
            }
        }

        public int Delay
        {
            get
            {
                return _moveClient.Delay;
            }
        }

        public void Connect()
        {
            if (Connected) return;
            _moveClient.Connect();
        }

        public void PauseTracking(bool paused)
        {
            if (paused && !_moveClient.IsPaused)
            {
                _moveClient.PauseTracking();
                return;
            }
            _moveClient.ResumeTracking();
        }

        public void Destroy()
        {
            UnsubscribeFromEvents();

            foreach (var move in Moves)
            {
                move.Destroy();
            }

            _moveClient.Destroy();
            _instance = null;
        }

        private void DetectMoves()
        {
            var connectedMoves = _moveClient.ConnectedMoves;

            var enumValues = Enum.GetValues(typeof (MoveClient.MoveID)).Cast<MoveClient.MoveID>();
            foreach (var moveID in enumValues.Where(moveID => moveID != MoveClient.MoveID.None && connectedMoves.HasFlag(moveID)))
            {
                Moves.Add(new Models.Move(moveID));
            }
        }

        private void SubscribeToEvents()
        {
            var adjustDelay = new AdjustDelayEvent();
            adjustDelay.Subscribe(OnAdjustDelay);

            _moveClient.ConnectedMovesUpdated += OnConnectedMovesUpdated;
        }

        private void UnsubscribeFromEvents()
        {
            var adjustDelay = new AdjustDelayEvent();
            adjustDelay.Unsubscribe(OnAdjustDelay);

            _moveClient.ConnectedMovesUpdated -= OnConnectedMovesUpdated;
        }

        private void OnAdjustDelay(AdjustDelayEventArgs args)
        {
            if (!args.Absolute)
            {
                args.Delay = _moveClient.Delay + args.Delay;
            }
            _moveClient.Delay = args.Delay;
        }

        private void OnConnectedMovesUpdated(MoveClient.MoveID moveID, MoveClient.ConnectionStatus connectionStatus)
        {
            if (connectionStatus == MoveClient.ConnectionStatus.Connected)
            {
                Moves.Add(new Models.Move(moveID));
            }
            else
            {
                Moves.Remove(Moves.Where(move => move.MoveID == moveID).FirstOrDefault());
            }
            _inputDetectedEvent.Publish(null);
        }
    }
}
