﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Timers;
using Maze.Extensions;
using Maze.Input.Wii.Converters;
using Microsoft.Xna.Framework;
using WiimoteLib;
using Timer = System.Timers.Timer;

namespace Maze.Input.Wii.Models.Extensions
{
    public class MotionPlus : WiiExtension
    {
        public event EventHandler Calibrated;

        private const int CalibrationTime = 100;
        private const int CalibrationTick = 10;
        private const int NoiseThreshold = 10;

        private readonly ToDegreesConverter _converter;
        private readonly CalibrationTimer _calibrationTimer;
        private List<float> _previousYawReadings;
        private DateTime _previousAutoCalibration;
        private TimeSpan _autoCalibrationQuarantine;

        private List<Vector3> _calibrationReadings;
        private double _calibrationTimeout;
        private Noise _yawNoise;
        private Noise _pitchNoise;
        private Noise _rollNoise;
        private float _yawBias;
        private float _pitchBias;
        private float _rollBias;
        private float _lastAngle = Int32.MaxValue;
        private Vector3 _previousAccelState = new Vector3(0, 0, 0);

        public MotionPlus(Wiimote wiimote) : base(wiimote)
        {
            _converter = new ToDegreesConverter();

            _calibrationTimer = new CalibrationTimer { Tick = CalibrationTick };
            _calibrationTimer.Elapsed += CalibrationTimerElapsed;

            Calibrate();
//            InitializeAutoCalibration();
        }

        protected MotionPlusState MotionPlusState
        {
            get
            {
                return Wiimote.WiimoteState.MotionPlusState;
            }
        }

        public Vector3 Acceleration
        {
            get
            {
                var currentValues = Wiimote.WiimoteState.AccelState.Values;
                var current = new Vector3((float)Math.Truncate(currentValues.X * 1000) / 1000, currentValues.Y, currentValues.Z);

                var xValue = current.X.AlmostEquals(_previousAccelState.X, 0.05f) ? _previousAccelState.X : current.X;
                var yValue = current.Y.AlmostEquals(_previousAccelState.Y, 0.05f) ? _previousAccelState.Y : current.Y;
                var zValue = current.Z.AlmostEquals(_previousAccelState.Z, 0.05f) ? _previousAccelState.Z : current.Z;

                _previousAccelState = current;

                return new Vector3(xValue, yValue, zValue);
            }
        }

        public float GetRotation()
        {
            var currentAngle = _converter.ConvertTo(Wiimote);

            if (currentAngle < 0)
            {
                currentAngle = 180 + (180 - Math.Abs(currentAngle));
            }

            if (_lastAngle == Int32.MaxValue)
            {
                _lastAngle = currentAngle;
            }

            if (_lastAngle - currentAngle > 250)
            {
                _lastAngle = _lastAngle - 360;
            }

            if (currentAngle - _lastAngle > 250)
            {
                _lastAngle = _lastAngle + 360;
            }

            _lastAngle = _lastAngle + (currentAngle - _lastAngle) * 0.05f;

            return _lastAngle;
        }

        public float GetRawYaw()
        {
            return Normalize(MotionPlusState.GyroState.Values.Yaw - _yawBias);
        }

        public float GetRawPitch()
        {
            return Normalize(MotionPlusState.GyroState.Values.Pitch - _pitchBias);
        }

        public float GetRawRoll()
        {
            return Normalize(MotionPlusState.GyroState.Values.Roll - _rollBias);
        }

        private float Normalize(float value)
        {
            return value.AlmostEquals(0, 2) ? 0 : value;
        }

        public void Calibrate()
        {
            Console.WriteLine("Calibrating MotionPlus!");
            
            _calibrationTimer.Stop();

            _yawBias = 0;
            _pitchBias = 0;
            _rollBias = 0;

            Wiimote.CalibrateMotionPlus();

            // Reset calibration values
            _calibrationReadings = new List<Vector3>();
            _calibrationTimeout = CalibrationTime;
            _yawNoise = new Noise(Int32.MaxValue, Int32.MinValue);
            _pitchNoise = new Noise(Int32.MaxValue, Int32.MinValue);
            _rollNoise = new Noise(Int32.MaxValue, Int32.MinValue);

            _calibrationTimer.Start();
        }

        private void CalibrationTimerElapsed(object sender, ElapsedEventArgs e)
        {
            // Determine if the Wiimote is moving
            if (IsMoving())
            {
                Console.WriteLine("Cancelling calibration because the Wiimote is moving");
                _calibrationTimer.Stop();

                Calibrate();
                return;
            }

            _calibrationReadings.Add(new Vector3(GetRawYaw(), GetRawPitch(), GetRawRoll()));

            _calibrationTimeout -= _calibrationTimer.Tick;
            if (_calibrationTimeout > 0) return;

            Console.WriteLine("Finishing calibration");
            _calibrationTimer.Stop();
            CalculateBias();
        }

        private bool IsMoving()
        {
            // Adjust noise readings
            _yawNoise.Adjust(GetRawYaw());
            _pitchNoise.Adjust(GetRawPitch());
            _rollNoise.Adjust(GetRawRoll());

            return Math.Abs(_yawNoise.Max - _yawNoise.Min) > NoiseThreshold
                || Math.Abs(_pitchNoise.Max - _pitchNoise.Min) > NoiseThreshold
                || Math.Abs(_rollNoise.Max - _rollNoise.Min) > NoiseThreshold;
        }

        private void CalculateBias()
        {
            float yawSum = 0;
            float pitchSum = 0;
            float rollSum = 0;

            foreach (var reading in _calibrationReadings)
            {
                yawSum += reading.X;
                pitchSum += reading.Y;
                rollSum += reading.Z;
            }

            _yawBias = yawSum / _calibrationReadings.Count;
            _pitchBias = pitchSum / _calibrationReadings.Count;
            _rollBias = rollSum / _calibrationReadings.Count;

            OnCalibrated();
        }

        private void InitializeAutoCalibration()
        {
            _previousAutoCalibration = DateTime.MinValue;
            _previousYawReadings = new List<float>();
            _autoCalibrationQuarantine = new TimeSpan(0, 0, 0, 10);

            var autoCalibrationTimer = new Timer(100) { AutoReset = true };
            autoCalibrationTimer.Elapsed += AutoCalibrationTimerElapsed;
            autoCalibrationTimer.Start();
        }

        private void AutoCalibrationTimerElapsed(object sender, ElapsedEventArgs e)
        {
            var currentYaw = GetRawYaw();
            
            _previousYawReadings.Add(currentYaw);

            if (_previousYawReadings.Count < 20)
            {
                return;
            }

            AttemptAutoCalibration();
            _previousYawReadings.Clear();
        }

        private void AttemptAutoCalibration()
        {
            var quarantineHasPassed = DateTime.Now.Subtract(_previousAutoCalibration) > _autoCalibrationQuarantine;
            
            if (quarantineHasPassed && IsYawValueStable())
            {
                Calibrate();
                _previousAutoCalibration = DateTime.Now;

                Console.WriteLine("Autocalibrated!");
            }
        }

        private bool IsYawValueStable()
        {
            var firstYawValue = _previousYawReadings.First();
            return _previousYawReadings.All(value => firstYawValue.AlmostEquals(value, 2f));
        }

        private void OnCalibrated()
        {
            if (Calibrated != null)
                Calibrated(this, EventArgs.Empty);
        }
    }
}
