﻿using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using Maze.Events;
using Maze.Input.Kinect;
using Microsoft.Xna.Framework.Input;
using KinectModel = Maze.Input.Kinect.Models.Kinect;

namespace Maze.Input.Wrapper
{
    public class KinectWrapper : UnpollableInputWrapper
    {
        private readonly RotateCameraEvent _rotateCameraEvent;
        private readonly KinectDetecter _kinectDetecter;
        private readonly List<Keys> _pushedKeys;
        private bool _gamePaused;

        public KinectWrapper()
        {
            _rotateCameraEvent = new RotateCameraEvent();
            _kinectDetecter = KinectDetecter.GetInstance();
            _pushedKeys = new List<Keys>();

            SubscribeToEvents();
        }

        public override List<string> Errors
        {
            get
            {
                var errorList = new List<string>();
                foreach (var kinect in _kinectDetecter.Kinects.Where(kinect => kinect.Errors.Count > 0))
                {
                    errorList.AddRange(kinect.Errors);
                }
                return errorList;
            }
        }

        public override bool IsGameplayAllowed
        {
            get
            {
                var gameplayDisallowed = _kinectDetecter.Kinects.Any(kinect => !kinect.GameplayAllowed);
                return !gameplayDisallowed;
            }
        }

        public override bool KeyPushed(Keys key)
        {
            if (_pushedKeys.Contains(key))
            {
                _pushedKeys.Remove(key);
                return true;
            }
            return false;
        }

        public override bool KeyDown(Keys key)
        {
            return false;
        }

        public override void Destroy()
        {
            UnsubscribeFromEvents();
        }

        private void SubscribeToEvents()
        {
            var gamePausedEvent = new GamePausedEvent();
            gamePausedEvent.Subscribe(OnGamePaused);

            _kinectDetecter.Kinects.CollectionChanged += KinectCollectionChanged;
            _kinectDetecter.Kinects.PublishInitialCollectionChanged();
        }

        private void UnsubscribeFromEvents()
        {
            var gamePausedEvent = new GamePausedEvent();
            gamePausedEvent.Subscribe(OnGamePaused);

            _kinectDetecter.Kinects.CollectionChanged -= KinectCollectionChanged;
        }

        private void KinectCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Reset || e.Action == NotifyCollectionChangedAction.Move)
                return;

            if (e.NewItems != null)
            {
                foreach (var kinect in from object item in e.NewItems select item as KinectModel)
                {
                    kinect.Activity += OnKinectActivity;
                    kinect.ActionDetected += OnActionDetected;
                }
            }

            if (e.OldItems != null)
            {
                foreach (var kinect in from object item in e.OldItems select item as KinectModel)
                {
                    kinect.Activity -= OnKinectActivity;
                    kinect.ActionDetected -= OnActionDetected;
                }
            }
        }

        private void OnActionDetected(KinectModel kinect, KinectModel.Actions action)
        {
            switch (action)
            {
                case KinectModel.Actions.SwipeLeft:
                case KinectModel.Actions.SwipeRight:
                    break;
                case KinectModel.Actions.LeftHandOverHead:
                    _pushedKeys.Add(Keys.Up);
                    break;
                case KinectModel.Actions.RightHandOverHead:
                    _pushedKeys.Add(Keys.Down);
                    break;
                case KinectModel.Actions.LeftHello:
                    _pushedKeys.Add(Keys.P);
                    break;
                case KinectModel.Actions.RightHello:
                    _pushedKeys.Add(Keys.Enter);
                    break;
            }
        }

        private void OnKinectActivity(KinectModel kinect)
        {
            if (_gamePaused) return;

            var args = new RotateCameraEventArgs { Sender = this, Absolute = true, Rotation = kinect.Angle };            
            _rotateCameraEvent.Publish(args);
        }

        private void OnGamePaused(GamePausedEventArgs args)
        {
            _gamePaused = args.Paused;
        }
    }
}
