﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using Maze.Events;
using Maze.Input.Move;
using Microsoft.Xna.Framework.Input;
using MoveModel = Maze.Input.Move.Models.Move;

namespace Maze.Input.Wrapper
{
    public class MoveWrapper : UnpollableInputWrapper
    {
        private readonly MoveDetecter _moveDetecter;
        private readonly Dictionary<Keys, MoveModel.Buttons> _keyMap;
        private readonly RotateCameraEvent _rotateCameraEvent;
        private bool _gamePaused;

        public MoveWrapper()
        {
            _moveDetecter = MoveDetecter.GetInstance();
            _keyMap = new Dictionary<Keys, MoveModel.Buttons>();
            _rotateCameraEvent = new RotateCameraEvent();

            InitializeKeyMap();
            SubscribeToEvents();
        }

        public override void Destroy()
        {
            UnsubscribeFromEvents();
            base.Destroy();
        }

        public override bool KeyPushed(Keys key)
        {
            if (!_keyMap.ContainsKey(key)) return false;

            var button = _keyMap[key];

            bool buttonPushed = false;
            foreach (var move in _moveDetecter.Moves.Where(move => move.ButtonPushed(button)))
            {
                switch (button)
                {
                    case MoveModel.Buttons.Select:
                        move.Reset();
                        break;
                    case MoveModel.Buttons.Triangle:
                        move.Calibrate();
                        break;
                }
                buttonPushed = true;
            }
            return buttonPushed;
        }

        public override bool KeyDown(Keys key)
        {
            var button = _keyMap[key];
            return _moveDetecter.Moves.Any(move => move.ButtonDown(button));
        }

        private void SubscribeToEvents()
        {
            var gamePausedEvent = new GamePausedEvent();
            gamePausedEvent.Subscribe(OnGamePaused);

            _moveDetecter.Moves.CollectionChanged += OnMoveCollectionChanged;
            _moveDetecter.Moves.PublishInitialCollectionChanged();
        }

        private void UnsubscribeFromEvents()
        {
            var gamePausedEvent = new GamePausedEvent();
            gamePausedEvent.Unsubscribe(OnGamePaused);

            _moveDetecter.Moves.CollectionChanged -= OnMoveCollectionChanged;
        }

        private void OnMoveCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Reset || e.Action == NotifyCollectionChangedAction.Move)
                return;

            if (e.NewItems != null)
            {
                foreach (var move in from object item in e.NewItems select item as MoveModel)
                {
                    move.Activity += OnMoveActivity;
                }
            }

            if (e.OldItems != null)
            {
                foreach (var move in from object item in e.OldItems select item as MoveModel)
                {
                    move.Activity -= OnMoveActivity;
                }
            }
        }

        private void OnMoveActivity(MoveModel move)
        {
            if (_gamePaused) return;

            var angle = Math.Atan2(move.Position.Y - move.HandlePosition.Y, move.Position.X - move.HandlePosition.X) * (180 / Math.PI);
            var args = new RotateCameraEventArgs { Sender = this, Absolute = true, Rotation = (float) -angle };

//            var args = new RotateCameraEventArgs { Sender = this, Absolute = true, Rotation = move.Orientation.X - 90 };
            _rotateCameraEvent.Publish(args);
        }

        private void OnGamePaused(GamePausedEventArgs args)
        {
            _gamePaused = args.Paused;
        }

        private void InitializeKeyMap()
        {
            _keyMap.Add(Keys.Enter, MoveModel.Buttons.Move);
            _keyMap.Add(Keys.Up, MoveModel.Buttons.Square);
            _keyMap.Add(Keys.Down, MoveModel.Buttons.Cross);
            _keyMap.Add(Keys.P, MoveModel.Buttons.Start);
            _keyMap.Add(Keys.F2, MoveModel.Buttons.Triangle);
            _keyMap.Add(Keys.F8, MoveModel.Buttons.Select);
        }
    }
}
