﻿using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using Maze.Events;
using Maze.Input.Wii;
using Maze.Input.Wii.Extensions;
using Microsoft.Xna.Framework.Input;
using WiimoteLib;

namespace Maze.Input.Wrapper
{
    public class WiiWrapper : UnpollableInputWrapper
    {
        private readonly WiiDetecter _wiiDetecter;
        private readonly Dictionary<Keys, WiiButton> _keyMap;
        private readonly RotateCameraEvent _rotateCameraEvent;
        private bool _gamePaused;

        public WiiWrapper()
        {
            _wiiDetecter = WiiDetecter.GetInstance();
            _keyMap = new Dictionary<Keys, WiiButton>();
            _rotateCameraEvent = new RotateCameraEvent();

            InitializeKeyMap();
            SubscribeToEvents();
        }

        public override void Destroy()
        {
            UnsubscribeFromEvents();
            base.Destroy();
        }

        public override bool KeyPushed(Keys key)
        {
            if (!_keyMap.ContainsKey(key)) return false;

            var button = Translate(key);
            return _wiiDetecter.Wiimotes.Any(wiimote => wiimote.KeyPushed(button));
        }

        public override bool KeyDown(Keys key)
        {
            if (!_keyMap.ContainsKey(key)) return false;

            var button = Translate(key);
            return _wiiDetecter.Wiimotes.Any(wiimote => wiimote.KeyDown(button));
        }

        private void SubscribeToEvents()
        {
            var gamePausedEvent = new GamePausedEvent();
            gamePausedEvent.Subscribe(OnGamePaused);

            _wiiDetecter.Wiimotes.CollectionChanged += OnWiimoteCollectionChanged;
            _wiiDetecter.Wiimotes.PublishInitialCollectionChanged();
        }

        private void UnsubscribeFromEvents()
        {
            var gamePausedEvent = new GamePausedEvent();
            gamePausedEvent.Unsubscribe(OnGamePaused);

            _wiiDetecter.Wiimotes.CollectionChanged -= OnWiimoteCollectionChanged;
        }

        private void OnWiimoteCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Reset || e.Action == NotifyCollectionChangedAction.Move)
                return;

            if (e.NewItems != null)
            {
                foreach (var wiimote in from object item in e.NewItems select item as Wiimote)
                {
                    wiimote.WiimoteChanged += OnWiimoteActivity;
                }
            }

            if (e.OldItems != null)
            {
                foreach (var wiimote in from object item in e.OldItems select item as Wiimote)
                {
                    wiimote.WiimoteChanged -= OnWiimoteActivity;
                }
            }
        }

        private void OnWiimoteActivity(object sender, WiimoteChangedEventArgs eventArgs)
        {
            if (_gamePaused) return;

            var wiimote = sender as Wiimote;

            if (!wiimote.HasMotionPlus()) return;

            var motionPlus = wiimote.GetMotionPlus();

            var args = new RotateCameraEventArgs { Sender = this, Absolute = true, Rotation = motionPlus.GetRotation() };
            _rotateCameraEvent.Publish(args);
        }

        private void OnGamePaused(GamePausedEventArgs args)
        {
            _gamePaused = args.Paused;
        }

        public override List<string> Errors
        {
            get
            {
                var errorList = new List<string>();

                if (!_wiiDetecter.HasMotionPlus())
                {
                    errorList.Add("Please connect a Wiimote with a Motion Plus extension.");
                }
                if (_wiiDetecter.HasBalanceBoard() && !IsPlayerBalanced())
                {
                    errorList.Add("Please stand in an upright position on the Balance Board");
                }

                return errorList;
            }
        }

        private WiiButton Translate(Keys key)
        {
            return _keyMap[key];
        }

        private bool IsPlayerBalanced()
        {
            var balanceBoard = _wiiDetecter.BalanceBoard;

            var xGravityIsCentered = balanceBoard.CenterOfGravity.X > -5 && balanceBoard.CenterOfGravity.X < 5;
            var yGravityIsCentered = balanceBoard.CenterOfGravity.Y > -5 && balanceBoard.CenterOfGravity.Y < 5;
            
            return balanceBoard.Weight > 20 && xGravityIsCentered && yGravityIsCentered;
        }

        private void InitializeKeyMap()
        {
            _keyMap.Add(Keys.Enter, WiiButton.A);
            _keyMap.Add(Keys.Left, WiiButton.Left);
            _keyMap.Add(Keys.Right, WiiButton.Right);
            _keyMap.Add(Keys.Up, WiiButton.Right);
            _keyMap.Add(Keys.Down, WiiButton.Left);
            _keyMap.Add(Keys.P, WiiButton.Home);
            _keyMap.Add(Keys.F2, WiiButton.One);
            _keyMap.Add(Keys.F3, WiiButton.Two);
        }
    }
}
