﻿using System;
using FlatRedBall;
using FlatRedBall.Graphics;
using Maze.Extensions;
using Maze.Helpers;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Maze.Models
{
    public class Ball : CollisionEntity
    {
        public const int MaxRotation = 1;
        private const int MaxGravity = 200;
        private const float RotationDecrementValue = .08f;

        private readonly RotatingCamera _rotatingCamera;
        private Vector3 _gravity;
        private bool _gravityEnabled;

        public Ball(string contentManager, Layer layer, RotatingCamera camera) : base(layer, contentManager)
        {
            _rotatingCamera = camera;

            Gravity = new Vector3(0, -MaxGravity, 0);
            Drag = 2f;
        }

        public Vector3 Gravity
        {
            get
            {
                return _gravity;
            }
            set
            {
                _gravity = value;

                if (IsGravityEnabled)
                {
                    Acceleration = Gravity;
                }
            }
        }

        public bool IsGravityEnabled
        {
            get
            {
                return _gravityEnabled;
            }
            set
            {
                _gravityEnabled = value;
                Acceleration = value ? Gravity : Vector3.Zero;
            }
        }

        public bool IsMoving
        {
            get
            {
                return IsGravityEnabled && this.IsMoving();
            }
        }

        public override void Initialize(bool addToManagers)
        {
            Texture = FlatRedBallServices.Load<Texture2D>(@"Content\Game\Ball", ContentManagerName);

            base.Initialize(addToManagers);
        }

        public override void Activity()
        {
            AdjustVelocity();
            AdjustRotation();           
//            Console.WriteLine("Velocity:" + Velocity + " | Acceleration: " + Acceleration + " | Rotation: " + RotationZVelocity);
        }

        public void AdjustGravity(RotatingCamera camera)
        {
            SetGravity(CalculateGravity(camera.RotationRad));
        }

        protected void SetGravity(Vector2 gravity)
        {
            Gravity = new Vector3(gravity, 0);
        }

        private void AdjustVelocity()
        {
            if (Velocity.X.AlmostEquals(0, 1f))
            {
                Velocity.X = 0;
            }
            if (Velocity.Y.AlmostEquals(0, 1f))
            {
                Velocity.Y = 0;
            }
        }

        private void AdjustRotation()
        {
            if (!Visible)
            {
                return;
            }
            var realMovementDirection = this.GetRealMovementDirection(_rotatingCamera);

            var rotation = BallHelper.GetProperRotation(realMovementDirection, _rotatingCamera.CameraAngle);
            if (rotation.AlmostEquals(0))
            {
                AdjustRotationToZero();
            }
            else
            {
                RotationZVelocity = rotation;
            }
        }

        private void AdjustRotationToZero()
        {
            var rotation = RotationZVelocity;

            if (rotation.AlmostEquals(0))
            {
                return;
            }

            if (rotation > 0)
            {
                rotation = Math.Max(0, rotation - RotationDecrementValue);
            } else if (rotation < 0)
            {
                rotation = Math.Min(0, rotation + RotationDecrementValue);
            }
            RotationZVelocity = rotation;
        }

        private static Vector2 CalculateGravity(float rotation)
        {
            var xGravity = Math.Sin(rotation) * MaxGravity;
            var yGravity = Math.Cos(rotation) * MaxGravity * -1;
            
            return new Vector2((float) xGravity, (float) yGravity);
        }
    }
}
