﻿using System;
using FlatRedBall.Graphics;

namespace Maze.Models
{
    public class Game : Entity
    {
        private static Game _gameInstance;
        private static int _instanceCount;

        public event EventHandler ScoreUpdated;
        public event EventHandler LevelUpdated;

        private readonly DateTime _gameStarted;
        private readonly LevelMap _levelMap;
        private int _score;
        private Level _level;

        protected Game(string contentManager) : base(contentManager)
        {
            _gameStarted = DateTime.UtcNow;
            _levelMap = new LevelMap();

            LevelNumber = 1;
        }

        public static Game GetInstance()
        {
            return _gameInstance ?? (_gameInstance = new Game("Game" + _instanceCount));
        }

        public static void ResetInstance()
        {
            if (_gameInstance == null) return;
            _gameInstance.Destroy();
            _gameInstance = null;
            _instanceCount++;
        }

        public TimeSpan TimeUsed
        {
            get
            {
                return DateTime.UtcNow - _gameStarted;
            }
        }

        public int Score
        {
            get
            {
                return _score;
            }
            private set
            {
                _score = value;
                OnScoreUpdated(EventArgs.Empty);
            }
        }

        public int LevelNumber
        {
            get;
            private set;
        }

        public Level Level
        {
            get
            {
                return _level;
            }
            private set
            {
                _level = value;
                OnLevelUpdated(EventArgs.Empty);
            }
        }

        public override void Activity()
        {
            Level.Activity();
        }

        public override void Destroy()
        {
            Level.Destroy();
        }

        public void AddScore(int score)
        {
            if (score <= 0)
            {
                throw new Exception("Cannot add a negative score");
            }
            Score += score;
        }

        public void IncrementLevel()
        {
            if (HasMoreLevels())
            {
                LevelNumber++;
            }
        }

        public bool HasMoreLevels()
        {
            return _levelMap.HasMoreLevels(Level);
        }

        public void LoadLevel(Layer layer)
        {
            LoadLevel(_levelMap.GetLevel(LevelNumber), layer);
        }

        protected void LoadLevel(LevelPath levelPath, Layer layer)
        {
            Level = LoadLevel(levelPath, ContentManagerName, layer);
            Level.MoveToCenter();
        }

        protected Level LoadLevel(LevelPath levelPath, string contentManager, Layer layer)
        {
            return new Level(levelPath, contentManager, layer);
        }

        protected virtual void OnScoreUpdated(EventArgs e)
        {
            var eventHandler = ScoreUpdated;
            if (eventHandler != null)
            {
                eventHandler(this, e);
            }
        }

        protected virtual void OnLevelUpdated(EventArgs e)
        {
            var eventHandler = LevelUpdated;
            if (eventHandler != null)
            {
                eventHandler(this, e);
            }
        }
    }
}
