﻿using System;
using FlatRedBall;
using FlatRedBall.Graphics;
using FlatRedBall.Math;
using Maze.Extensions;
using Maze.Helpers;
using Maze.Input;
using Microsoft.Xna.Framework;

namespace Maze.Models.HUDs
{
    public abstract class BaseHUD : Entity
    {
        protected static Camera Camera;

        protected InputManager InputManager;
        protected PositionedObjectList<Text> Texts;

        protected BaseHUD(string contentManagerName) : base(contentManagerName)
        {
            InputManager = InputManager.GetInstance();
            InitializeCamera();
        }

        protected BitmapFont Font
        {
            get;
            set;
        }

        protected Text ErrorText
        {
            get;
            private set;
        }

        public override void Initialize(bool addToManagers)
        {
            Texts = new PositionedObjectList<Text>();

            Font = new BitmapFont(
                "Content/Fonts/calibri_48",
                "Content/Fonts/calibri_48.fnt",
                FlatRedBallServices.GlobalContentManager);

            base.Initialize(addToManagers);
        }

        public override void AddToManager(Layer layer)
        {
            InitializeErrorText(layer);

            base.AddToManager(layer);
        }

        public override void Activity()
        {
            ErrorText.SetPixelPerfectScale(Camera);

            base.Activity();
        }

        public override void Destroy()
        {
            TextManager.RemoveText(Texts);
            base.Destroy();
        }

        protected Text CreateText(Vector3 position, Layer layer)
        {
            var text = TextManager.AddText("", layer);
            text.Font = Font;
            text.Position = position;

            text.Red = 1f;
            text.Green = 1f;
            text.Blue = 1f;

            text.SetPixelPerfectScale(layer);
            Texts.Add(text);

            return text;
        }

        protected void UpdateErrorText(string text)
        {
            if (ErrorText.DisplayText.Equals(text))
            {
                return;
            }

            ErrorText.DisplayText = text;

            var xValue = Globals.CenterX - ErrorText.TextCenter();
            if (Math.Abs(xValue - 0) < .1f)
            {
                xValue = Globals.CenterX;
            }

            ErrorText.Position = new Vector3(xValue, Globals.CenterY, 0);
        }

        protected void DestroyCamera()
        {
            if (Camera == null) return;

            CameraHelper.DestroyCamera(Camera);
            Camera = null;
        }

        private void InitializeCamera()
        {
            if (Camera != null) return;

            Camera = new Camera(ContentManagerName) { DrawsWorld = false };
            Camera.SetupCamera();
            SpriteManager.Cameras.Add(Camera);
        }

        private void InitializeErrorText(Layer layer)
        {
            ErrorText = TextManager.AddText("", layer);
            ErrorText.HorizontalAlignment = HorizontalAlignment.Center;
            ErrorText.VerticalAlignment = VerticalAlignment.Center;
            ErrorText.Font = Font;
            ErrorText.Red = 1f;
            ErrorText.Green = 1f;
            ErrorText.Blue = 1f;

            Texts.Add(ErrorText);
        }
    }
}
