﻿using System;
using System.Collections.Generic;
using System.Linq;
using FlatRedBall;
using FlatRedBall.Graphics;
using FlatRedBall.Math;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Maze.Models.HUDs
{
    public class GameHUD : BaseHUD
    {
        private const string ScoreString = "Score: ";
        private const string LevelString = "Level: ";

        private readonly Game _game;
        private PositionedObjectList<BaseHUD> _inputSpecificHUD;
        private Texture2D _leaningTexture;
        private Texture2D _tooFarTexture;
        private Texture2D _tooCloseTexture;
        private Texture2D _tooCloseToBodyTexture;

        public GameHUD(Game game, string contentManagerName) : base(contentManagerName)
        {
            _game = game;

            Initialize(false);
            AddToManager(Camera.Layer);
        }

        public Text ScoreText
        {
            get;
            private set;
        }

        public Text LevelText
        {
            get;
            private set;
        }

        public Text ScoreDisplay
        {
            get;
            private set;
        }

        public Text LevelDisplay
        {
            get;
            private set;
        }

        public Sprite Indicator
        {
            get;
            private set;
        }

        public override void Initialize(bool addToManagers)
        {
            base.Initialize(addToManagers);

            SubscribeToEvents();
            _inputSpecificHUD = new PositionedObjectList<BaseHUD>();

            _leaningTexture = FlatRedBallServices.Load<Texture2D>(@"Content\Game\Wii\Leaning", ContentManagerName);
            _tooFarTexture = FlatRedBallServices.Load<Texture2D>(@"Content\Game\Kinect\HandsTooFarAway", ContentManagerName);
            _tooCloseTexture = FlatRedBallServices.Load<Texture2D>(@"Content\Game\Kinect\HandsTooClose", ContentManagerName);
            _tooCloseToBodyTexture = FlatRedBallServices.Load<Texture2D>(@"Content\Game\Kinect\HandsTooCloseToBody", ContentManagerName);
            
            InitializeInputSpecificHUD();
        }

        public override void AddToManager(Layer layer)
        {
            InitializeScoreText(layer);
            InitializeLevelText(layer);
            InitializeScoreDisplay(layer);
            InitializeLevelDisplay(layer);
            InitializeIndicators(layer);

            ScoreDisplay.DisplayText = _game.Score.ToString();
            LevelDisplay.DisplayText = _game.Level.ToString();

            base.AddToManager(layer);
        }

        public override void Activity()
        {
            var errors = InputManager.GetErrors();
            if (errors.Count > 0)
            {
                var error = errors.Last();
                UpdateErrorText(error);
            }
            else
            {
                ErrorText.DisplayText = "";
            }
            DisplayErrorIndicator(errors);

            foreach (var hud in _inputSpecificHUD)
            {
                hud.Activity();
            }

            base.Activity();
        }

        private void DisplayErrorIndicator(List<string> errors)
        {
            if (errors.Count == 0)
            {
                Indicator.Visible = false;
                return;
            }

            var error = errors.Last();

            if (error.Contains("upright position"))
            {
                ShowIndicator(_leaningTexture);
                return;
            }
            if (error.Contains("hands are too close to eachother"))
            {
                ShowIndicator(_tooCloseTexture);
                return;
            }
            if (error.Contains("hands are too far away"))
            {
                ShowIndicator(_tooFarTexture);
                return;
            }
            if (error.Contains("hands are too close to your body"))
            {
                ShowIndicator(_tooCloseToBodyTexture);
                return;
            }
            Indicator.Visible = false;
        }

        private void ShowIndicator(Texture2D texture)
        {
            Indicator.Texture = texture;
            Indicator.Visible = true;
        }

        public override void Destroy()
        {
            foreach(var hud in _inputSpecificHUD)
            {
                hud.Destroy();
            }

            SpriteManager.RemoveSprite(Indicator);

            UnsubscribeFromEvents();

            DestroyCamera();
            base.Destroy();
        }

        private void InitializeScoreText(Layer layer)
        {
            ScoreText = TextManager.AddText(ScoreString, layer);
            ScoreText.Font = Font;
            ScoreText.Position = new Vector3(50, Globals.ViewportHeight - 50, 0);

            ScoreText.Red = 1f;
            ScoreText.Green = 1f;
            ScoreText.Blue = 1f;

            ScoreText.SetPixelPerfectScale(layer);

            Texts.Add(ScoreText);
        }

        private void InitializeLevelText(Layer layer)
        {
            LevelText = TextManager.AddText(LevelString, layer);
            LevelText.Font = Font;
            LevelText.Position = new Vector3(50, Globals.ViewportHeight - 100, 0);

            LevelText.Red = 1f;
            LevelText.Green = 1f;
            LevelText.Blue = 1f;

            LevelText.SetPixelPerfectScale(layer);

            Texts.Add(LevelText);
        }

        private void InitializeScoreDisplay(Layer layer)
        {
            ScoreDisplay = TextManager.AddText("", layer);
            ScoreDisplay.Font = Font;
            ScoreDisplay.Position = new Vector3(190, Globals.ViewportHeight - 50, 0);

            ScoreDisplay.Red = 1f;
            ScoreDisplay.Green = 1f;
            ScoreDisplay.Blue = 1f;

            ScoreDisplay.SetPixelPerfectScale(layer);

            Texts.Add(ScoreDisplay);
        }

        private void InitializeLevelDisplay(Layer layer)
        {
            LevelDisplay = TextManager.AddText("", layer);
            LevelDisplay.Font = Font;
            LevelDisplay.Position = new Vector3(190, Globals.ViewportHeight - 100, 0);

            LevelDisplay.Red = 1f;
            LevelDisplay.Green = 1f;
            LevelDisplay.Blue = 1f;

            LevelDisplay.SetPixelPerfectScale(layer);

            Texts.Add(LevelDisplay);
        }

        private void InitializeIndicators(Layer layer)
        {
            Indicator = SpriteManager.AddSprite(_leaningTexture, layer);
            Indicator.PixelSize = .5f;
            Indicator.Position = new Vector3(150, Globals.CenterY, 0);
            Indicator.Visible = false;
        }

        private void InitializeInputSpecificHUD()
        {
            foreach(var hud in _inputSpecificHUD)
            {
                hud.Destroy();
            }
            _inputSpecificHUD.Clear();

            if (InputManager.DetectedInput.HasFlag(Input.Input.Wii))
            {
                _inputSpecificHUD.Add(new WiiHUD(ContentManagerName));
            }
            if (InputManager.DetectedInput.HasFlag(Input.Input.Kinect))
            {
                _inputSpecificHUD.Add(new KinectHUD(ContentManagerName));
            }
            if (InputManager.DetectedInput.HasFlag(Input.Input.Move))
            {
                _inputSpecificHUD.Add(new MoveHUD(ContentManagerName));
            }
        }

        private void SubscribeToEvents()
        {
            InputManager.DetectedInputChanged += InputDetected;

            _game.ScoreUpdated += ScoreUpdated;
            _game.LevelUpdated += LevelUpdated;
        }

        private void UnsubscribeFromEvents()
        {
            InputManager.DetectedInputChanged -= InputDetected;

            _game.ScoreUpdated -= ScoreUpdated;
            _game.LevelUpdated -= LevelUpdated;
        }

        private void InputDetected(object sender, EventArgs e)
        {
            InitializeInputSpecificHUD();
        }

        private void ScoreUpdated(object sender, EventArgs e)
        {
            ScoreDisplay.DisplayText = _game.Score.ToString();
        }

        private void LevelUpdated(object sender, EventArgs e)
        {
            LevelDisplay.DisplayText = _game.Level.ToString();
        }
    }
}
