﻿using System;
using FlatRedBall;
using FlatRedBall.Graphics;
using FlatRedBall.Math;
using Maze.Events;
using Maze.Input.Kinect.Events;
using Maze.Input.Wrapper;
using Microsoft.Research.Kinect.Nui;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Point = Microsoft.Xna.Framework.Point;

namespace Maze.Models.HUDs
{
    public class KinectHUD : BaseHUD
    {
        private const int MaxSkeletonX = 320;
        private const int MaxSkeletonY = 240;

        private readonly KinectWrapper _kinectWrapper;
        private readonly PositionedObjectList<Sprite> _sprites;
        private Sprite _headSprite;
        private Sprite _leftHandSprite;
        private Sprite _rightHandSprite;
        private Sprite _videoSprite;
        private Vector3 _initialSpritePosition;

        public KinectHUD(string contentManagerName) : base(contentManagerName)
        {
            _kinectWrapper = InputManager.GetInputWrapper<KinectWrapper>();
            _sprites = new PositionedObjectList<Sprite>();

            SubscribeToEvents();

            Initialize(false);
            AddToManager(Camera.Layer);
        }

        protected Text SteeringWheelAngle
        {
            get;
            set;
        }

        protected Text CameraTilt
        {
            get;
            set;
        }

        protected Text LeftHandTrackingStatus
        {
            get;
            set;
        }

        protected Text RightHandTrackingStatus
        {
            get;
            set;
        }

        public override void AddToManager(Layer layer)
        {
            InitializeSteeringWheelAngleDisplay(layer);
            InitializeCameraTiltDisplay(layer);
            InitializeLeftHandTrackingStatusDisplay(layer);
            InitializeRightHandTrackingStatusDisplay(layer);
            InitializeOverlayBackground(layer);
            InitializeVideoOverlay(layer);
            InitializeSkeletonOverlay(layer);

            base.AddToManager(layer);
        }

        public override void Destroy()
        {
            UnsubscribeFromEvents();
            SpriteManager.RemoveSpriteList(_sprites);
            base.Destroy();
        }

        private void SubscribeToEvents()
        {
            var skeletonProcessedEvent = new SkeletonProcessedEvent();
            skeletonProcessedEvent.Subscribe(OnSkeletonReady);

            var videoProcessedEvent = new VideoProcessedEvent();
            videoProcessedEvent.Subscribe(OnVideoReady);

            var cameraTiltAdjustedEvent = new CameraTiltAdjustedEvent();
            cameraTiltAdjustedEvent.Subscribe(OnTiltAdjusted);

            var cameraRotatedEvent = new CameraRotatedEvent();
            cameraRotatedEvent.Subscribe(OnCameraRotated);
        }

        private void UnsubscribeFromEvents()
        {
            var skeletonProcessedEvent = new SkeletonProcessedEvent();
            skeletonProcessedEvent.Unsubscribe(OnSkeletonReady);

            var videoProcessedEvent = new VideoProcessedEvent();
            videoProcessedEvent.Unsubscribe(OnVideoReady);

            var cameraTiltAdjustedEvent = new CameraTiltAdjustedEvent();
            cameraTiltAdjustedEvent.Unsubscribe(OnTiltAdjusted);

            var cameraRotatedEvent = new CameraRotatedEvent();
            cameraRotatedEvent.Unsubscribe(OnCameraRotated);
        }

        private void OnCameraRotated(CameraRotatedEventArgs args)
        {
            SteeringWheelAngle.DisplayText = "Angle: " + args.Angle;
        }

        private void OnTiltAdjusted(CameraTiltAdjustedEventArgs args)
        {
            CameraTilt.DisplayText = "Camera Tilt: " + args.Tilt;
        }

        private void OnVideoReady(VideoProcessedEventArgs args)
        {
            _videoSprite.Texture = args.Frame;
        }

        private void UpdateStatusTexts(SkeletonData skeleton)
        {
            if (skeleton == null || skeleton.TrackingState != SkeletonTrackingState.Tracked)
            {
                LeftHandTrackingStatus.DisplayText = "";
                RightHandTrackingStatus.DisplayText = "";
                return;
            }

            String leftHandText;
            switch (skeleton.Joints[JointID.HandLeft].TrackingState)
            {
                case JointTrackingState.Tracked:
                    leftHandText = "Tracked";
                    break;
                case JointTrackingState.Inferred:
                    leftHandText = "Inferred";
                    break;
                default:
                    leftHandText = "Not Tracked";
                    break;
            }
            LeftHandTrackingStatus.DisplayText = "Left Hand: " + leftHandText;

            String rightHandText;
            switch (skeleton.Joints[JointID.HandRight].TrackingState)
            {
                case JointTrackingState.Tracked:
                    rightHandText = "Tracked";
                    break;
                case JointTrackingState.Inferred:
                    rightHandText = "Inferred";
                    break;
                default:
                    rightHandText = "Not Tracked";
                    break;
            }
            RightHandTrackingStatus.DisplayText = "Right Hand: " + rightHandText;
        }

        private void OnSkeletonReady(SkeletonProcessedEventArgs args)
        {
            UpdateStatusTexts(args.Skeleton);

            if (args.Skeleton == null)
            {
                SetSpritePosition(_headSprite);
                SetSpritePosition(_leftHandSprite);
                SetSpritePosition(_rightHandSprite);
                return;
            }

            var skeleton = args.Skeleton;
            if (skeleton.TrackingState != SkeletonTrackingState.Tracked) return;

            var runtime = args.Kinect.Runtime;
            SetSpritePosition(_headSprite, skeleton.Joints[JointID.Head], runtime);
            SetSpritePosition(_leftHandSprite, skeleton.Joints[JointID.HandLeft], runtime);
            SetSpritePosition(_rightHandSprite, skeleton.Joints[JointID.HandRight], runtime);
        }

//        private static void SetSpritePosition(Sprite sprite, Joint joint)
//        {
//            var scaledJoint = joint.ScaleTo(MaxSkeletonX, MaxSkeletonY);
//
//            var xPosition = scaledJoint.Position.X;
//            var yPosition = MaxSkeletonY - scaledJoint.Position.Y;
//            sprite.Position = new Vector3(xPosition, yPosition, 0);
//        }

        private void SetSpritePosition(Sprite sprite)
        {
            sprite.Position = GetInitialPosition();
        }

        private void SetSpritePosition(Sprite sprite, Joint joint, Runtime runtime)
        {
            sprite.Position = joint.TrackingState == JointTrackingState.Tracked
                  ? GetPosition(joint, runtime)
                  : GetInitialPosition();
        }

        private Vector3 GetInitialPosition()
        {
            return _initialSpritePosition;
        }

        private Vector3 GetPosition(Joint joint, Runtime runtime)
        {
            var jointPosition = GetDisplayPosition(runtime, joint);

            var xPosition = jointPosition.X / 2;
            var yPosition = MaxSkeletonY - jointPosition.Y / 2;
            return new Vector3(xPosition, yPosition, 0);
        }

        private static Point GetDisplayPosition(Runtime nui, Joint joint)
        {
            float depthX, depthY;
            nui.SkeletonEngine.SkeletonToDepthImage(joint.Position, out depthX, out depthY);
            depthX = Math.Max(0, Math.Min(depthX * 320, 320));  //convert to 320, 240 space
            depthY = Math.Max(0, Math.Min(depthY * 240, 240));  //convert to 320, 240 space

            int colorX, colorY;
            var iv = new ImageViewArea();
            // only ImageResolution.Resolution640x480 is supported at this point
            nui.NuiCamera.GetColorPixelCoordinatesFromDepthPixel(ImageResolution.Resolution640x480, iv, (int)depthX, (int)depthY, (short)0, out colorX, out colorY);

            return new Point(colorX, colorY);
        }

        private void InitializeSteeringWheelAngleDisplay(Layer layer)
        {
            SteeringWheelAngle = TextManager.AddText("", layer);
            SteeringWheelAngle.Font = Font;
            SteeringWheelAngle.Position = new Vector3(10, 270, 0);

            SteeringWheelAngle.Red = 1f;
            SteeringWheelAngle.Green = 1f;
            SteeringWheelAngle.Blue = 1f;

            SteeringWheelAngle.SetPixelPerfectScale(layer);

            Texts.Add(SteeringWheelAngle);
        }

        private void InitializeCameraTiltDisplay(Layer layer)
        {
            CameraTilt = TextManager.AddText("", layer);
            CameraTilt.Font = Font;
            CameraTilt.Position = new Vector3(10, 310, 0);

            CameraTilt.Red = 1f;
            CameraTilt.Green = 1f;
            CameraTilt.Blue = 1f;

            CameraTilt.SetPixelPerfectScale(layer);

            Texts.Add(CameraTilt);
        }

        private void InitializeRightHandTrackingStatusDisplay(Layer layer)
        {
            RightHandTrackingStatus = TextManager.AddText("", layer);
            RightHandTrackingStatus.Font = Font;
            RightHandTrackingStatus.Position = new Vector3(10, 350, 0);

            RightHandTrackingStatus.Red = 1f;
            RightHandTrackingStatus.Green = 1f;
            RightHandTrackingStatus.Blue = 1f;

            RightHandTrackingStatus.SetPixelPerfectScale(layer);

            Texts.Add(RightHandTrackingStatus);
        }

        private void InitializeLeftHandTrackingStatusDisplay(Layer layer)
        {
            LeftHandTrackingStatus = TextManager.AddText("", layer);
            LeftHandTrackingStatus.Font = Font;
            LeftHandTrackingStatus.Position = new Vector3(10, 390, 0);

            LeftHandTrackingStatus.Red = 1f;
            LeftHandTrackingStatus.Green = 1f;
            LeftHandTrackingStatus.Blue = 1f;

            LeftHandTrackingStatus.SetPixelPerfectScale(layer);

            Texts.Add(LeftHandTrackingStatus);
        }

        private void InitializeOverlayBackground(Layer layer)
        {
            var rectangleTexture = FlatRedBallServices.Load<Texture2D>(@"Content\GUI\Kinect\Rectangle", ContentManagerName);
            var rectangle = SpriteManager.AddSprite(rectangleTexture, layer);
            rectangle.ScaleX = MaxSkeletonX / 2;
            rectangle.ScaleY = MaxSkeletonY / 2;
            rectangle.Position = new Vector3(rectangle.ScaleX, rectangle.ScaleY, 0);
            _sprites.Add(rectangle);
        }

        private void InitializeVideoOverlay(Layer layer)
        {
            _videoSprite = new Sprite { ScaleX = MaxSkeletonX / 2, ScaleY = MaxSkeletonY / 2 };
            _videoSprite.Position = new Vector3(_videoSprite.ScaleX, _videoSprite.ScaleY, 0);
            SpriteManager.AddSprite(_videoSprite);
            SpriteManager.AddToLayer(_videoSprite, layer);
            _sprites.Add(_videoSprite);
        }

        private void InitializeSkeletonOverlay(Layer layer)
        {
            _initialSpritePosition = new Vector3(-100, -100, 0);

            var headTexture = FlatRedBallServices.Load<Texture2D>(@"Content\GUI\Kinect\Head", ContentManagerName);
            _headSprite = SpriteManager.AddSprite(headTexture, layer);
            _headSprite.PixelSize = .5f;
            _headSprite.Position = _initialSpritePosition;
            _sprites.Add(_headSprite);

            var leftHandTexture = FlatRedBallServices.Load<Texture2D>(@"Content\GUI\Kinect\LeftHand", ContentManagerName);
            _leftHandSprite = SpriteManager.AddSprite(leftHandTexture, layer);
            _leftHandSprite.PixelSize = .5f;
            _leftHandSprite.Position = _initialSpritePosition;
            _sprites.Add(_leftHandSprite);

            var rightHandTexture = FlatRedBallServices.Load<Texture2D>(@"Content\GUI\Kinect\RightHand", ContentManagerName);
            _rightHandSprite = SpriteManager.AddSprite(rightHandTexture, layer);
            _rightHandSprite.PixelSize = .5f;
            _rightHandSprite.Position = _initialSpritePosition;
            _sprites.Add(_rightHandSprite);
        }
    }
}
