﻿using System;
using System.Collections.Generic;
using System.Linq;
using FlatRedBall;
using FlatRedBall.Graphics;
using FlatRedBall.ManagedSpriteGroups;
using FlatRedBall.Math.Geometry;
using Maze.Helpers;
using Maze.Models.Items;
using Microsoft.Xna.Framework;

namespace Maze.Models
{
    public class Level
    {
        public const string MapSuffix = ".scnx";
        public const string CollisionSuffix = ".shcx";

        public Level(LevelPath levelPath, string contentManager, Layer layer = null)
        {
            LevelPath = levelPath;

            Map = LevelHelper.LoadScene(LevelPath + MapSuffix, contentManager, layer);
            Collision = LevelHelper.LoadShapeCollection(LevelPath + CollisionSuffix, contentManager, layer);

            LoadFinishPoints(contentManager, layer);
            LoadItems(contentManager, layer);

            CollisionVisible = Globals.IsCollisionsVisible;
        }

        public LevelPath LevelPath
        {
            get;
            private set;
        }

        public int LevelNumber
        {
            get
            {
                return LevelPath.LevelNumber;
            }
        }

        public Scene Map
        {
            get;
            private set;
        }

        public ShapeCollection Collision
        {
            get;
            private set;
        }

        public List<SpriteGrid> SpriteGrids
        {
            get
            {
                return Map.SpriteGrids;
            }
        }

        public SpriteList Sprites
        {
            get
            {
                return Map.Sprites;
            }
        }

        public FinishPoints FinishPoints
        {
            get;
            private set;
        }

        public ItemList ItemList
        {
            get;
            private set;
        }

        public Sprite BallPlaceholder
        {
            get
            {
                return Sprites.FindByName("Ball");
            }
        }

        public bool CollisionVisible
        {
            set
            {
                Collision.Visible = value;
            }
        }

        public void Activity()
        {
            Map.ManageAll();
        }

        public void Destroy()
        {
            SpriteManager.RemoveScene(Map, true);
            Collision.RemoveFromManagers();
            FinishPoints.Destroy();
            ItemList.Destroy();
        }

        public override string ToString()
        {
            return LevelNumber.ToString();
        }

        public void MoveToCenter()
        {
            Move(Globals.CenterOfScreen);
        }

        public void Move(Vector2 coordinate)
        {
            foreach (var spriteGrid in SpriteGrids)
            {
                LevelHelper.MoveSpriteGrid(spriteGrid, coordinate);
            }
            var grid = SpriteGrids.FirstOrDefault();
            LevelHelper.MoveShapeCollection(Collision, grid, coordinate);
            LevelHelper.MoveSprites(Sprites, grid, coordinate);
            LevelHelper.MoveFinishPoints(FinishPoints, grid, coordinate);
            LevelHelper.MoveItems(ItemList, grid, coordinate);
        }

        private void LoadFinishPoints(string contentManager, Layer layer)
        {
            FinishPoints = new FinishPoints();

            var finishPlaceholders = Sprites.FindSpritesWithNameContaining("Finish");
            foreach (var finishPlaceholder in finishPlaceholders)
            {
                var finish = new Finish(layer, contentManager);
                finish.PositionSpriteUsingPlaceholderAsReference(finishPlaceholder);
                finish.Scale(12);

                FinishPoints.Add(finish);

                finishPlaceholder.Visible = false;
            }
        }

        private void LoadItems(string contentManager, Layer layer)
        {
            ItemList = new ItemList();

            var itemPlaceholders = Sprites.FindSpritesWithNameContaining("Item");
            foreach (var itemPlaceholder in itemPlaceholders)
            {
                Item item;

                var number = RandomNumber(10, 40) % 4;
                switch (number)
                {
                    case 0:
                        item = new Lemon(layer, contentManager);
                        break;
                    case 1:
                        item = new Banana(layer, contentManager);
                        break;
                    case 2:
                        item = new Apple(layer, contentManager);
                        break;
                    default:
                        item = new Strawberry(layer, contentManager);
                        break;
                }
                item.PositionSpriteUsingPlaceholderAsReference(itemPlaceholder);
                item.Scale(12);

                ItemList.Add(item);
                itemPlaceholder.Visible = false;
            }
        }

        private static readonly Random Random = new Random();
        private static readonly object SyncLock = new object();

        private static int RandomNumber(int min, int max)
        {
            lock (SyncLock)
            {
                return Random.Next(min, max);
            }
        }
    }
}
