﻿using System;
using System.Timers;
using FlatRedBall;
using FlatRedBall.Graphics;
using FlatRedBall.Math.Geometry;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Maze.Models.Menu
{
    public class Button : Entity
    {
        public event EventHandler Clicked;

        private Camera _camera;
        private Texture2D _buttonTexture;
        private Texture2D _hoveredButtonTexture;
        private Texture2D _clickedButtonTexture;
        private AxisAlignedRectangle _collision;

        private readonly string _originalText;
        private Timer _clickedTimer;

        public Button(string text, string contentManagerName, Layer layer = null) : base(contentManagerName)
        {
            _originalText = text;

            Initialize(false);
            AddToManager(layer);
        }

        public bool Selected
        {
            get;
            set;
        }

        public Sprite Sprite
        {
            get;
            private set;
        }

        public AxisAlignedRectangle Collision
        {
            get
            {
                return _collision;
            }
        }

        public Text Text
        {
            get;
            private set;
        }

        protected bool IsClicked
        {
            get;
            set;
        }

        public override void Initialize(bool addToManagers)
        {
            _camera = SpriteManager.Camera;

            _buttonTexture = FlatRedBallServices.Load<Texture2D>("Content/Menu/Button", ContentManagerName);
            _hoveredButtonTexture = FlatRedBallServices.Load<Texture2D>("Content/Menu/HoveredButton", ContentManagerName);
            _clickedButtonTexture = FlatRedBallServices.Load<Texture2D>("Content/Menu/ClickedButton", ContentManagerName);

            base.Initialize(addToManagers);
        }

        public override void AddToManager(Layer layer)
        {
            base.AddToManager(layer);

            InitializeNormalSprite(layer);
            InitializeText(layer);
            InitializeCollision(layer);
        }

        public override void Activity()
        {
            DisplayButtonState();
        }

        public override void Destroy()
        {
            base.Destroy();

            SpriteManager.RemoveSprite(Sprite);
            TextManager.RemoveText(Text);
            ShapeManager.Remove(_collision);
        }

        public void WasClicked()
        {
            _clickedTimer = new Timer(100);
            _clickedTimer.Elapsed += ClickedTimerElapsed;
            _clickedTimer.AutoReset = false;
            _clickedTimer.Start();

            IsClicked = true;
        }

        protected virtual void OnClicked(EventArgs e)
        {
            var eventHandler = Clicked;

            if (eventHandler != null)
            {
                eventHandler(this, e);
            }
        }

        public void PositionButton(Vector2 startingPoint)
        {
            PositionButton(startingPoint, Vector2.Zero);
        }

        public void PositionButton(Vector2 startingPoint, Vector2 offset)
        {
            PositionButton(startingPoint, offset, 0);
        }

        public void PositionButton(Vector2 startingPoint, Vector2 offset, int zValue)
        {
            if (offset == Vector2.Zero)
            {
                Position = new Vector3(startingPoint, zValue);
                return;
            }
            Position = new Vector3(startingPoint + offset, zValue);
        }

        private void InitializeNormalSprite(Layer layer)
        {
            Sprite = SpriteManager.AddSprite(_buttonTexture, layer);
            Sprite.AttachTo(this, false);
            Sprite.ScaleX = 0.5f * Sprite.Texture.Width / _camera.PixelsPerUnitAt(Sprite.Z);
            Sprite.ScaleY = 0.5f * Sprite.Texture.Height / _camera.PixelsPerUnitAt(Sprite.Z);
        }

        private void InitializeText(Layer layer)
        {
            Text = TextManager.AddText("", layer);
            Text.AttachTo(this, false);

            Text.DisplayText = _originalText;

            Text.HorizontalAlignment = HorizontalAlignment.Center;
            Text.VerticalAlignment = VerticalAlignment.Center;

            Text.Red = 0f;
            Text.Green = 0f;
            Text.Blue = 0f;

            Text.Font = new BitmapFont(
                "Content/Fonts/calibri_48", 
                "Content/Fonts/calibri_48.fnt", 
                FlatRedBallServices.GlobalContentManager);
            Text.SetPixelPerfectScale(_camera);
            Text.RelativeY = .1f;
        }

        private void InitializeCollision(Layer layer)
        {
            _collision = new AxisAlignedRectangle { ScaleX = Sprite.ScaleX, ScaleY = Sprite.ScaleY };
            _collision.AttachTo(this, false);

            ShapeManager.AddAxisAlignedRectangle(_collision);
            ShapeManager.AddToLayer(_collision, layer);
            _collision.Visible = false;
        }

        private void DisplayButtonState()
        {
            if (IsClicked)
            {
                ShowClickedButton();
            }
            else if (Selected)
            {
                ShowHoveredButton();
            } else
            {
                ShowNormalButton();
            }
        }

        private void ShowNormalButton()
        {
            Sprite.Texture = _buttonTexture;
        }

        private void ShowHoveredButton()
        {
            Sprite.Texture = _hoveredButtonTexture;
        }

        private void ShowClickedButton()
        {
            Sprite.Texture = _clickedButtonTexture;
        }

        private void ClickedTimerElapsed(object sender, ElapsedEventArgs e)
        {
            IsClicked = false;
            OnClicked(EventArgs.Empty);
        }
    }
}
