﻿using System;
using FlatRedBall;
using Maze.Extensions;
using Maze.Helpers;
using Microsoft.Xna.Framework;

namespace Maze.Models
{
    public class RotatingCamera : Camera
    {
        public delegate void RotationChangedEventHandler(RotatingCamera sender, EventArgs e);

        public event RotationChangedEventHandler RotationChanged;

        public const float RotationTick = 1;

        public enum Direction
        {
            Left, Right
        }

        public RotatingCamera(string contentManagerName) : base(contentManagerName)
        {
            Setup();
        }

        public RotatingCamera(string contentManagerName, int width, int height) : base(contentManagerName, width, height)
        {
            Setup();
        }

        public Sprite CameraSprite
        {
            get;
            private set;
        }

        public float Rotation
        {
            get
            {
                return MathHelper.ToDegrees(CameraSprite.RotationZ);
            }
        }

        public float RotationRad
        {
            get
            {
                return CameraSprite.RotationZ;
            }
        }

        public DirectionHelper.CameraAngle CameraAngle
        {
            get
            {
                return CameraHelper.GetCameraAngle(this);
            }
        }

        public float XAxisRotation
        {
            get
            {
                var cameraRotation = Rotation;

                float xAxisRotation;
                if (cameraRotation >= 90)
                {
                    xAxisRotation = cameraRotation - 90;
                }
                else
                {
                    xAxisRotation = 360 - 90 + cameraRotation;
                }
                return xAxisRotation;
            }
        }

        public void Activity()
        {
            UpVector = RotationMatrix.Up;
        }

        public void Destroy()
        {
            SpriteManager.RemoveSprite(CameraSprite);
        }

        public void Rotate(Direction direction)
        {
            if (direction == Direction.Left)
            {
                CameraSprite.RotationZ -= MathHelper.ToRadians(RotationTick);
            }
            if (direction == Direction.Right)
            {
                CameraSprite.RotationZ += MathHelper.ToRadians(RotationTick);
            }

            OnRotationChanged();
//            Console.WriteLine("Adjusted the camera's rotation. Value is " + Rotation + " degrees (" + XAxisRotation + " degrees measured from the X-axis)");
        }

        public void Rotate(float angle, bool relativeToCurrentRotation = false)
        {
            if (relativeToCurrentRotation)
            {
                ChangeRelativeRotation(angle);
            }
            else
            {
                ChangeAbsoluteRotation(angle);
            }
            OnRotationChanged();
        }

        private void ChangeAbsoluteRotation(float angle)
        {
            CameraSprite.RotationZ = MathHelper.ToRadians(angle);
        }

        private void ChangeRelativeRotation(float angle)
        {
            if (angle < -360 || angle > 360)
            {
                throw new Exception("Cannot rotate the camera more than 360 degrees in either direction.");
            }
            SetRotation(Rotation + angle);
        }

        private void SetRotation(float angle)
        {
            CameraSprite.RotationZ = MathHelper.ToRadians(angle);
        }

        public void ResetRotation()
        {
            SetRotation(0);
        }

        public void Setup()
        {
            Enable2DMode();
            InitializeCameraSprite();
            AttachTo(CameraSprite, true);
        }

        private void OnRotationChanged()
        {
            if (RotationChanged != null)
            {
                RotationChanged(this, new EventArgs());
            }
        }

        private void Enable2DMode()
        {
            this.SetupCamera();
        }

        private void InitializeCameraSprite()
        {
            CameraSprite = SpriteManager.AddSprite(@"Content\redball");
            CameraSprite.Position = new Vector3(Globals.CenterOfScreen, 0);
            CameraSprite.PixelSize = .5f;
            CameraSprite.Visible = false;
            SpriteManager.AddToLayer(CameraSprite, Layer);
        }
    }
}
