﻿using System;
using System.Linq;
using FlatRedBall;
using FlatRedBall.Instructions;
using Maze.Events;
using Maze.Input;
using Maze.Input.Kinect.Events;
using Maze.Input.Move.Events;
using Maze.Input.Wii;
using Maze.Input.Wii.Extensions;
using Maze.Input.Wrapper;
using Maze.Models;
using Maze.Models.HUDs;
using Maze.Screens.Pause;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Game = Maze.Models.Game;

namespace Maze.Screens
{
    public class GameScreen : RotatingCameraScreen
    {
        private readonly InputManager _inputManager;
        private Texture2D _borderTexture;
        private Texture2D _borderRedTexture;
        private Texture2D _mazeBoardTexture;
        private GamePausedScreen _pauseScreen;
        private LevelCompletedScreen _levelCompletedScreen;
        private bool _gamePaused;

        public GameScreen() : base("GameScreen")
        {
            _inputManager = InputManager.GetInstance();
        }

        public Game Game
        {
            get;
            private set;
        }

        public Level Level
        {
            get
            {
                return Game.Level;
            }
        }

        public Ball Ball
        {
            get;
            private set;
        }

        public Sprite MazeBorder
        {
            get;
            private set;
        }

        public Sprite MazeBoard
        {
            get;
            private set;
        }

        public GameHUD HUD
        {
            get;
            private set;
        }

        protected bool IsGameStopped
        {
            get
            {
                return InstructionManager.IsEnginePaused;
            }
        }

        private bool IsGamePaused
        {
            get
            {
                return _gamePaused;
            }
            set
            {
                _gamePaused = value;
                var gamePausedEvent = new GamePausedEvent();
                gamePausedEvent.Publish(new GamePausedEventArgs() { Paused = value });
            }
        }

        private bool IsLevelFinished
        {
            get;
            set;
        }

        public override void Initialize(bool addToManagers)
        {
            LoadBackground(@"Menu\Background");
            _borderTexture = LoadTexture(@"Game\MazeBorder");
            _borderRedTexture = LoadTexture(@"Game\MazeBorderRed");
            _mazeBoardTexture = LoadTexture(@"Game\MazeBoard");

//            EnableFiltering(false);
//            Camera.EnableCullMode(false);

            base.Initialize(addToManagers);
        }

        public override void AddToManagers()
        {
            base.AddToManagers();

            InitializeMazeBorder();
            InitializeMazeBoard();

            Game = Game.GetInstance();
            Game.LoadLevel(RotatingCamera.Layer);
            
            InitializeBall();
            InitializeHUD();

            SubscribeToEvents();
        }

        public override void Destroy()
        {
            UnsubscribeFromEvents();

            Game.Destroy();
            Ball.Destroy();
            HUD.Destroy();

            SpriteManager.RemoveSpriteList(mSprites);

            base.Destroy();
        }

        public override void Activity(bool firstTimeCalled)
        {
            if (firstTimeCalled)
            {
                ContinueGame();
            }

            HandleLevelCompleteMode();
            HandlePauseMode();

            if (IsGamePaused || IsLevelFinished)
            {
                base.Activity(firstTimeCalled);
                return;
            }

            HandleInputActivity();
            HUD.Activity();

            HandleGameplayAllowed();

            Game.Activity();
            Ball.Activity();

            PerformCollisions();
            HandlePlayerErrors();

            base.Activity(firstTimeCalled);
        }

        private void HandleInputActivity()
        {
            if (_inputManager.KeyPushed(Keys.F4))
            {
                IsLevelFinished = true;
            }

            if (_inputManager.HasPollableWrapper)
            {
                foreach (var wrapper in _inputManager.Wrappers.PollableWrappers.Where(wrapper => wrapper.HasRotationActivity))
                {
                    RotateCamera(wrapper.RotationAmount, true);
                }
            }

            if (_inputManager.DetectedInput.HasFlag(Input.Input.Wii))
            {
                if (_inputManager.KeyPushed(Keys.F2))
                {
                    var wiiDetecter = WiiDetecter.GetInstance(); ;
                    foreach (var wiimote in wiiDetecter.Wiimotes.Where(wiimote => wiimote.HasMotionPlus()))
                    {
                        wiimote.GetMotionPlus().Calibrate();
                    }
                }
            }

            if (_inputManager.DetectedInput.HasFlag(Input.Input.Kinect))
            {
                if (_inputManager.KeyPushed(Keys.Add))
                {
                    var adjustCameraTiltEvent = new AdjustCameraTiltEvent();
                    adjustCameraTiltEvent.Publish(new AdjustCameraTiltEventArgs() { Absolute = false, Tilt = 5});
                }
                if (_inputManager.KeyPushed(Keys.Subtract))
                {
                    var adjustCameraTiltEvent = new AdjustCameraTiltEvent();
                    adjustCameraTiltEvent.Publish(new AdjustCameraTiltEventArgs() { Absolute = false, Tilt = -5 });
                }
            }

            if (_inputManager.DetectedInput.HasFlag(Input.Input.Move))
            {
                if (_inputManager.KeyPushed(Keys.Add))
                {
                    var adjustDelayEvent = new AdjustDelayEvent();
                    adjustDelayEvent.Publish(new AdjustDelayEventArgs() { Absolute = false, Delay = 5 });
                }
                if (_inputManager.KeyPushed(Keys.Subtract))
                {
                    var adjustDelayEvent = new AdjustDelayEvent();
                    adjustDelayEvent.Publish(new AdjustDelayEventArgs() { Absolute = false, Delay = -5 });
                }
            }
        }

        private void HandleLevelCompleteMode()
        {
            if (!IsLevelFinished && LevelFinished())
            {
                IsLevelFinished = true;
                PauseGame();
            }

            if (IsLevelFinished)
            {
                DisplayLevelCompleteScreen();
            }
        }

        private void HandlePauseMode()
        {
            if (!IsGamePaused && UserPushedPause())
            {
                IsGamePaused = true;
                PauseGame();
            }

            if (IsGamePaused)
            {
                DisplayPauseScreen();
            }
        }

        private void HandleGameplayAllowed()
        {
            var gameplayDisallowed = _inputManager.Wrappers.Any(wrapper => wrapper.IsActive && !wrapper.IsGameplayAllowed);

            if (!IsGameStopped && gameplayDisallowed)
            {
                PauseGame();
            }
            if (IsGameStopped && !gameplayDisallowed)
            {
                ContinueGame();
            }

            if (IsGameStopped)
            {
                return;
            }
        }

        private void PerformCollisions()
        {
            Ball.Collision.CollideAgainstBounce(Level.Collision, 0, 1, 1);
            PerformItemCollision();
        }

        private void PerformItemCollision()
        {
            foreach (var item in Level.ItemList.Where(item => Ball.Collision.CollideAgainst(item.Collision)))
            {
                item.HasCollided();
            }
        }

        private void HandlePlayerErrors()
        {
            var errors = InputManager.GetErrors();
            MazeBorder.Texture = errors.Count > 0 ? _borderRedTexture : _borderTexture;
        }

        private bool UserPushedPause()
        {
//            return InputManager.KeyPushed(Keys.Escape) || InputManager.KeyPushed(Keys.P);
            return _inputManager.KeyPushed(Keys.P);
        }

        private bool LevelFinished()
        {
            return Ball.Collision.CollideAgainst(Level.FinishPoints.Collision);
        }

        private void DisplayLevelCompleteScreen()
        {
            if (_levelCompletedScreen != null) return;

            var layer = SpriteManager.Cameras.Last().AddLayer();
            _levelCompletedScreen = LoadPopup<LevelCompletedScreen>(layer);
            _levelCompletedScreen.LevelFinished += OnLevelFinished;
            _levelCompletedScreen.GameCompleted += OnGameCompleted;
        }

        private void DisplayPauseScreen()
        {
            if (_pauseScreen != null) return;

            var layer = SpriteManager.Cameras.Last().AddLayer();
            _pauseScreen = LoadPopup<GamePausedScreen>(layer);
            _pauseScreen.GameContinuing += OnGameContinuing;
            _pauseScreen.GameShuttingDown += OnGameShuttingDown;
        }

        private void DestroyPauseScreen()
        {
            var camera = SpriteManager.Cameras.Last();
            camera.RemoveLayer(_pauseScreen.Layer);
            _pauseScreen.GameContinuing -= OnGameContinuing;
            _pauseScreen.GameShuttingDown -= OnGameShuttingDown;
            _pauseScreen = null;
        }

        private void OnGameContinuing(object sender, EventArgs e)
        {
            DestroyPauseScreen();
            IsGamePaused = false;
            ContinueGame();
        }

        private void OnGameShuttingDown(object sender, Type screenToLoad)
        {
            DestroyPauseScreen();
            IsActivityFinished = true;
            MoveToScreen(screenToLoad.FullName);
        }

        private void OnLevelFinished(object sender, EventArgs e)
        {
            IsActivityFinished = true;
            MoveToScreen(typeof(GameScreen).FullName);
            ContinueGame();
        }

        private void OnGameCompleted(object sender, EventArgs e)
        {
            IsActivityFinished = true;
            MoveToScreen(typeof(GameCompletedScreen).FullName);
        }

        private void OnCameraRotationChanged(RotatingCamera sender, EventArgs e)
        {
            if (!IsGameStopped)
            {
                Ball.AdjustGravity(sender);
            }
        }

        private void InitializeBall()
        {
            Ball = new Ball(ContentManagerName, RotatingCamera.Layer, RotatingCamera);
            Ball.PositionSpriteUsingPlaceholderAsReference(Level.BallPlaceholder);
            Level.BallPlaceholder.Visible = false;

            Ball.Scale(12);
            Ball.IsGravityEnabled = true;
        }

        private void InitializeMazeBorder()
        {
            MazeBorder = SpriteManager.AddSprite(_borderTexture, RotatingCamera.Layer);
            MazeBorder.PixelSize = .5f;
            MazeBorder.Position = new Vector3(Globals.CenterOfScreen, 0);
            mSprites.Add(MazeBorder);
        }

        private void InitializeMazeBoard()
        {
            MazeBoard = SpriteManager.AddSprite(_mazeBoardTexture, RotatingCamera.Layer);
            MazeBoard.PixelSize = .5f;
            MazeBoard.Position = new Vector3(Globals.CenterOfScreen, 0);
            mSprites.Add(MazeBoard);
        }

        private void InitializeHUD()
        {
            HUD = new GameHUD(Game, ContentManagerName);
        }

        private void SubscribeToEvents()
        {
            RotatingCamera.RotationChanged += OnCameraRotationChanged;
        }

        private void UnsubscribeFromEvents()
        {
            RotatingCamera.RotationChanged -= OnCameraRotationChanged;
        }

        private static void PauseGame()
        {
            if (!InstructionManager.IsEnginePaused)
            {
                InstructionManager.PauseEngine();
            }
        }

        private static void ContinueGame()
        {
            if (InstructionManager.IsEnginePaused)
            {
                InstructionManager.UnpauseEngine();
            }
        }
    }
}