﻿using System;
using System.Linq;
using FlatRedBall.Graphics;
using Maze.Extensions;
using Maze.Models.Menu;
using Maze.Screens.Menu;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using Game = Maze.Models.Game;

namespace Maze.Screens.Pause
{
    public class GamePausedScreen : PausedScreen
    {
        public delegate void GameShuttingDownEventHandler(object sender, Type screenToLoad);

        public event EventHandler GameContinuing;
        public event GameShuttingDownEventHandler GameShuttingDown;

        protected ButtonCollection Buttons;

        public GamePausedScreen() : base("GamePaused")
        {
        }

        public Text PausedText
        {
            get;
            set;
        }

        public Button ContinueGameButton
        {
            get;
            private set;
        }

        public Button NewGameButton
        {
            get;
            private set;
        }

        public Button ExitToMenuButton
        {
            get;
            private set;
        }

        public override void AddToManagers()
        {
            base.AddToManagers();

            Buttons = new ButtonCollection();

            InitializePausedText(Layer);
            InitializeContinueGameButton(Layer);
            InitializeNewGameButton(Layer);
            InitializeExitToMenuButton(Layer);
        }

        public override void Activity(bool firstTimeCalled)
        {
            base.Activity(firstTimeCalled);

            if (firstTimeCalled && Buttons.Count > 0)
            {
                Buttons.SelectButton(0);
            }

            if (InputManager.KeyPushed(Keys.P))
            {
                ContinueGame();
            }

            MouseActivity();
            KeyboardActivity();

            for (int i = Buttons.Count - 1; i > -1; i--)
            {
                Buttons[i].Activity();
            }
        }

        public override void Destroy()
        {
            base.Destroy();

            while (Buttons.Count != 0)
            {
                Buttons.Last.Destroy();
            }
        }

        private void MouseActivity()
        {
            foreach (var button in Buttons.Where(button => InputManager.Mouse.IsOn(button.Collision)))
            {
                Buttons.SelectButton(button);
                if (IsMouseClicked())
                {
                    button.WasClicked();
                }
            }
        }

        private void KeyboardActivity()
        {
            if (InputManager.KeyPushed(Keys.Up))
            {
                Buttons.SelectPreviousButton();
            }
            if (InputManager.KeyPushed(Keys.Down))
            {
                Buttons.SelectNextButton();
            }
            if (InputManager.KeyPushed(Keys.Enter))
            {
                Buttons.SelectedButton.WasClicked();
            }
        }

        private bool IsMouseClicked()
        {
            return InputManager.Mouse.ButtonPushed(FlatRedBall.Input.Mouse.MouseButtons.LeftButton);
        }

        private void ContinueGame()
        {
            IsActivityFinished = true;
            OnEvent(GameContinuing, EventArgs.Empty);
        }

        private void InitializePausedText(Layer layer)
        {
            PausedText = TextManager.AddText("Game Paused", layer);
            PausedText.Font = Font;
            PausedText.SetPixelPerfectScale(layer);

            PausedText.Position = new Vector3(Globals.CenterX - PausedText.TextCenter(), Globals.ViewportHeight - 200, ZPosition);
            PausedText.Red = 1f;
            PausedText.Green = 1f;
            PausedText.Blue = 1f;

            Texts.Add(PausedText);
        }

        private void InitializeContinueGameButton(Layer layer)
        {
            ContinueGameButton = new Button("Continue", ContentManagerName, layer);
            ContinueGameButton.PositionButton(Globals.CenterOfScreen, new Vector2(0, 100), ZPosition);
            Buttons.Add(ContinueGameButton);

            ContinueGameButton.Clicked += ContinueGameButtonClicked;
        }

        private void InitializeNewGameButton(Layer layer)
        {
            NewGameButton = new Button("New Game", ContentManagerName, layer);
            NewGameButton.PositionButton(Globals.CenterOfScreen, Vector2.Zero, ZPosition);
            Buttons.Add(NewGameButton);

            NewGameButton.Clicked += NewGameButtonClicked;
        }

        private void InitializeExitToMenuButton(Layer layer)
        {
            ExitToMenuButton = new Button("Exit to Menu", ContentManagerName, layer);
            ExitToMenuButton.PositionButton(Globals.CenterOfScreen, new Vector2(0, -100), ZPosition);
            Buttons.Add(ExitToMenuButton);

            ExitToMenuButton.Clicked += ExitToMenuButtonClicked;
        }

        private void ContinueGameButtonClicked(object sender, EventArgs e)
        {
            ContinueGame();
        }

        private void NewGameButtonClicked(object sender, EventArgs e)
        {
            IsActivityFinished = true;
            Game.ResetInstance();
            OnGameShuttingDown(typeof(GameScreen));
        }

        private void ExitToMenuButtonClicked(object sender, EventArgs e)
        {
            IsActivityFinished = true;
            OnGameShuttingDown(typeof(MainMenu));
        }

        private void OnGameShuttingDown(Type screentoload)
        {
            var handler = GameShuttingDown;
            if (handler != null) handler(this, screentoload);
        }
    }
}
